# Copyright The Orthority Contributors.
#
# This file is part of Orthority.
#
# Orthority is free software: you can redistribute it and/or modify it under the terms of the GNU
# Affero General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# Orthority is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Orthority.
# If not, see <https://www.gnu.org/licenses/>.

"""Parameter file IO and conversions."""
from __future__ import annotations

import csv
import json
import logging
import os
import warnings
from abc import ABC, abstractmethod
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor
from csv import Dialect, DictReader, Sniffer
from io import StringIO
from os import PathLike
from pathlib import Path
from typing import Any, IO, Sequence

import cv2
import fsspec
import numpy as np
import rasterio as rio
import yaml
from fsspec.core import OpenFile
from rasterio.crs import CRS
from rasterio.errors import CRSError as RioCrsError
from rasterio.transform import RPC
from rasterio.warp import transform
from tqdm.std import tqdm

from orthority import utils
from orthority.enums import CameraType, CsvFormat
from orthority.errors import CrsError, CrsMissingError, ParamError
from orthority.exif import Exif

logger = logging.getLogger(__name__)

# TODO: define custom file types for e.g. str | Path | OpenFile | IO[str] once sphinx bug with
#  linking to external type defs in __init__ type hints is fixed

_opt_frame_schema = {
    CameraType.pinhole: ['sensor_size', 'cx', 'cy'],
    # fmt: off
    CameraType.opencv: [
        'sensor_size', 'cx', 'cy', 'k1', 'k2', 'p1', 'p2', 'k3', 'k4', 'k5', 'k6', 's1', 's2',
        's3', 's4', 'tx', 'ty',
    ],
    # fmt: on
    CameraType.brown: ['sensor_size', 'cx', 'cy', 'k1', 'k2', 'p1', 'p2', 'k3'],
    CameraType.fisheye: ['sensor_size', 'cx', 'cy', 'k1', 'k2', 'k3', 'k4'],
}
"""Schema of valid optional parameters for each frame camera type."""

_default_lla_crs = CRS.from_epsg(4979)
"""Default CRS for geographic camera coordinates."""

_default_tqdm_kwargs = dict(
    bar_format='{l_bar}{bar}|{n_fmt}/{total_fmt} files [{elapsed}<{remaining}]',
    dynamic_ncols=True,
    leave=True,
)
"""Default progress bar kwargs."""


def _read_osfm_int_param(json_dict: dict) -> dict[str, dict[str, Any]]:
    """Read interior parameters from an OpenDroneMap / OpenSfM JSON dictionary."""

    def parse_json_param(json_param: dict, cam_id: str) -> dict[str, Any]:
        """Validate & convert the given JSON dictionary for a single camera."""
        int_param = {}
        for req_key in ['projection_type', 'width', 'height']:
            if req_key not in json_param:
                raise ParamError(f"'{req_key}' is missing for camera '{cam_id}'.")

        # set 'cam_type' from 'projection_type'
        proj_type = json_param.pop('projection_type').lower()
        try:
            int_param['cam_type'] = CameraType.from_odm(proj_type)
        except ValueError:
            raise ParamError(f"Unsupported projection type '{proj_type}'.")

        im_size = (json_param.pop('width'), json_param.pop('height'))
        int_param['im_size'] = im_size

        # read focal length(s) (JSON values are normalised by max of sensor width & height)
        if 'focal' in json_param:
            int_param['focal_len'] = json_param.pop('focal')
        elif 'focal_x' in json_param and 'focal_y' in json_param:
            focal_x, focal_y = json_param.pop('focal_x'), json_param.pop('focal_y')
            int_param['focal_len'] = focal_x if focal_x == focal_y else (focal_x, focal_y)
        else:
            raise ParamError(
                f"'focal', or 'focal_x' and 'focal_y' are missing for camera '{cam_id}'."
            )

        # rename c_x->cx & c_y->cy
        for from_key, to_key in zip(['c_x', 'c_y'], ['cx', 'cy']):
            if from_key in json_param:
                int_param[to_key] = json_param.pop(from_key)

        # validate any remaining optional params, update param_dict & return
        err_keys = set(json_param.keys()).difference(_opt_frame_schema[int_param['cam_type']])
        if len(err_keys) > 0:
            raise ParamError(f"Unsupported parameter(s) {err_keys} for camera '{cam_id}'.")
        int_param.update(**json_param)
        return int_param

    # extract cameras section if json_dict is from an OpenSfM reconstruction.json file
    if isinstance(json_dict, list) and len(json_dict) == 1 and 'cameras' in json_dict[0]:
        json_dict = json_dict[0]['cameras']

    # parse each set of interior parameters
    int_param_dict = {}
    for cam_id, json_param in json_dict.items():
        cam_id = cam_id[3:] if cam_id.startswith('v2 ') else cam_id
        int_param_dict[cam_id] = parse_json_param(json_param.copy(), cam_id)

    return int_param_dict


def _create_exif_cam_id(exif: Exif) -> str:
    """Return a camera ID string for the given Exif object."""
    prop_list = [prop for prop in [exif.make, exif.model, exif.serial] if prop]
    cam_id = ' '.join(prop_list)
    if len(cam_id) == 0:
        cam_id = 'unknown'
    return cam_id


def _read_exif_int_param(exif: Exif) -> dict[str, dict[str, Any]]:
    """Read interior parameters from an Exif object."""
    # TODO: might there be cases where XMP tags CalibratedFocalLength, CalibratedOpticalCenter* are
    #  present but not DewarpData, and are better than equiv EXIF tags?
    if exif.dewarp:
        if len(exif.dewarp) != 9 or not any(exif.dewarp) or not exif.tag_im_size:
            logger.debug(f"Cannot interpret dewarp data for '{exif.filename}'.")
        else:
            # construct brown camera parameters from dewarp data and return
            cam_dict = dict(
                cam_type=CameraType.brown,
                im_size=exif.im_size,
                focal_len=tuple(exif.dewarp[:2]),
                sensor_size=(float(exif.tag_im_size[0]), float(exif.tag_im_size[1])),
            )
            cam_dict['cx'], cam_dict['cy'] = tuple(
                (np.array(exif.dewarp[2:4]) / max(exif.tag_im_size)).tolist()
            )
            dist_params = dict(zip(['k1', 'k2', 'p1', 'p2', 'k3'], exif.dewarp[-5:]))
            cam_dict.update(**dist_params)
            return {_create_exif_cam_id(exif): cam_dict}

    # construct pinhole camera parameters
    cam_dict = dict(cam_type=CameraType.pinhole, im_size=exif.im_size)
    if exif.focal_len and exif.sensor_size:
        # prefer using focal length and sensor size directly
        cam_dict['focal_len'] = exif.focal_len
        cam_dict['sensor_size'] = exif.sensor_size
    elif exif.focal_len_35:
        logger.debug(
            f"Approximating the focal length for '{exif.filename}' from the 35mm equivalent."
        )
        if exif.sensor_size:
            # scale 35mm focal length to actual focal length in mm, assuming "35mm" = 36mm max
            # sensor dimension
            cam_dict['focal_len'] = max(exif.sensor_size) * exif.focal_len_35 / 36.0
            cam_dict['sensor_size'] = exif.sensor_size
        else:
            # normalise 35mm focal length assuming "35mm" = 36 mm max sensor dimension, and find
            # a normalised sensor size in same units, assuming square pixels
            cam_dict['focal_len'] = exif.focal_len_35 / 36.0
    else:
        raise ParamError(
            f"No focal length & sensor size, or 35mm focal length tags in '{exif.filename}'."
        )

    return {_create_exif_cam_id(exif): cam_dict}


def _read_exif_ext_param(
    exif: Exif, crs: str | CRS, lla_crs: str | CRS
) -> dict[str, dict[str, Any]]:
    """Read exterior parameters from an Exif object."""
    if not exif.lla:
        raise ParamError(f"No latitude, longitude & altitude tags in '{exif.filename}'.")
    if not exif.rpy:
        raise ParamError(f"No camera / gimbal roll, pitch & yaw tags in '{exif.filename}'.")
    rpy = tuple(np.radians(exif.rpy).tolist())
    opk = _rpy_to_opk(rpy, exif.lla, crs, lla_crs=lla_crs)
    xyz = transform(lla_crs, crs, [exif.lla[1]], [exif.lla[0]], [exif.lla[2]])
    xyz = tuple([coord[0] for coord in xyz])
    ext_param = dict(xyz=xyz, opk=opk, camera=_create_exif_cam_id(exif))
    return {exif.filename: ext_param}


def read_oty_int_param(file: str | PathLike | OpenFile | IO[str]) -> dict[str, dict[str, Any]]:
    """
    Read interior parameters for one or more cameras from an :doc:`Orthority interior parameter
    file <../file_formats/oty_int>`.

    :param file:
        File to read.  Can be a path or URI string, an :class:`~fsspec.core.OpenFile` object or a
        file object, opened in text mode (``'rt'``).
    """
    with utils.Open(file, 'rt') as f:
        yaml_dict = yaml.safe_load(f)

    def parse_yaml_param(yaml_param: dict, cam_id: str = None) -> dict[str, Any]:
        """Validate & convert the given YAML dictionary for a single camera."""
        # test required keys for all cameras
        for req_key in ['type', 'im_size', 'focal_len']:
            if req_key not in yaml_param:
                raise ParamError(f"'{req_key}' is missing for camera '{cam_id}'.")

        # convert type -> cam_type
        cam_type = yaml_param.pop('type').lower()
        if cam_type not in CameraType.__members__ or cam_type == 'rpc':
            raise ParamError(f"Unsupported frame camera type '{cam_type}'.")
        int_param = dict(cam_type=CameraType(cam_type))

        int_param['im_size'] = tuple(yaml_param.pop('im_size'))

        # pop known legacy keys not supported by Camera.__init__
        yaml_param.pop('name', None)

        # set focal_len & sensor_size
        focal_len = yaml_param.pop('focal_len')
        int_param['focal_len'] = tuple(focal_len) if isinstance(focal_len, list) else focal_len
        if 'sensor_size' in yaml_param:
            int_param['sensor_size'] = tuple(yaml_param.pop('sensor_size'))

        # validate any remaining distortion params, update param_dict & return
        err_keys = set(yaml_param.keys()).difference(_opt_frame_schema[int_param['cam_type']])
        if len(err_keys) > 0:
            raise ParamError(f"Unsupported parameter(s) {err_keys} for camera '{cam_id}'.")
        int_param.update(**yaml_param)
        return int_param

    # flatten if in original simple-ortho format
    if 'camera' in yaml_dict:
        warnings.warn(
            "Support for the 'config.yaml' format is deprecated and will be removed in future. "
            "Please switch to the Orthority interior parameter format.",
            category=DeprecationWarning,
        )
        yaml_dict = yaml_dict['camera']

    # convert to nested dict if in flat format
    first_value = next(iter(yaml_dict.values()))
    if not isinstance(first_value, dict):
        cam_id = yaml_dict['name'] if 'name' in yaml_dict else 'unknown'
        yaml_dict = {cam_id: yaml_dict}

    # parse each set of interior parameters
    int_param_dict = {}
    for cam_id, yaml_param in yaml_dict.items():
        int_param_dict[cam_id] = parse_yaml_param(yaml_param, cam_id)
    return int_param_dict


def read_osfm_int_param(file: str | PathLike | OpenFile | IO[str]) -> dict[str, dict[str, Any]]:
    """
    Read interior parameters for one or more cameras, from an OpenDroneMap :file:`cameras.json` or
    OpenSfM :file:`reconstruction.json` file.

    See the :doc:`format documentation <../file_formats/opensfm>` for supported camera models.

    :param file:
        File to read.  Can be a path or URI string, an :class:`~fsspec.core.OpenFile` object or a
        file object, opened in text mode (``'rt'``).
    """
    with utils.Open(file, 'rt') as f:
        json_dict = json.load(f)

    return _read_osfm_int_param(json_dict)


def read_exif_int_param(
    file: str | PathLike | OpenFile | rio.DatasetReader,
) -> dict[str, dict[str, Any]]:
    """
    Read interior parameters for a camera from an image file with EXIF / XMP tags.

    See the :doc:`format documentation <../file_formats/exif_xmp>` for required tags.

    :param file:
        File to read.  Can be a path or URI string, an :class:`~fsspec.core.OpenFile` object in
        binary mode (``'rb'``), or a dataset reader.
    """
    return _read_exif_int_param(Exif(file))


def read_im_rpc_param(
    files: Sequence[str | PathLike | OpenFile | rio.DatasetReader],
    progress: bool | dict = False,
) -> dict[str, dict[str, Any]]:
    """
    Read RPC camera parameters from :doc:`image file(s) with RPC tags / sidecar file(s)
    <../file_formats/image_rpc>`.

    :param files:
        File(s) to read as a tuple of paths or URI strings, :class:`~fsspec.core.OpenFile`
        objects in binary mode (``'rb'``), or dataset readers.
    :param progress:
        Whether to display a progress bar monitoring the portion of files read.  Can be set to a
        dictionary of arguments for a custom `tqdm <https://tqdm.github.io/docs/tqdm/>`_ bar.
    """

    def _read_im_rpc_param(
        file: str | PathLike | OpenFile | rio.DatasetReader,
    ) -> dict[str, dict[str, Any]]:
        """Read RPC camera parameters from an image file."""
        filename = utils.get_filename(file)
        with utils.suppress_no_georef(), rio.Env(GDAL_NUM_THREADS='ALL_CPUS'), utils.OpenRaster(
            file, 'r'
        ) as im:
            # TODO: what is the speed of this for a large remote image?  does it just read the
            #  metadata, or the whole image?
            im_size = (im.width, im.height)
            rpc: RPC = im.rpcs

        if rpc is None:
            raise ParamError(f"No RPC parameters found in '{filename}'.")
        rpc_param = dict(cam_type=CameraType.rpc, im_size=im_size, rpc=rpc.to_dict())
        return {filename: rpc_param}

    # read RPC params in a thread pool, populating rpc_param_dict in same order as files
    rpc_param_dict = {}
    with ThreadPoolExecutor() as executor:
        futures = [executor.submit(_read_im_rpc_param, file) for file in files]

        # create / set up progress bar
        if progress is True:
            progress = tqdm(futures, **_default_tqdm_kwargs)
        elif progress is False:
            progress = tqdm(futures, disable=True, leave=False)
        else:
            progress = tqdm(futures, **progress)

        for future in progress:
            rpc_param_dict.update(**future.result())

    return rpc_param_dict


def read_oty_rpc_param(file: str | PathLike | OpenFile | IO[str]) -> dict[str, dict[str, Any]]:
    """
    Read RPC parameters for one or more cameras from an :doc:`Orthority RPC file
    <../file_formats/oty_rpc>`.

    :param file:
        File to read.  Can be a path or URI string, an :class:`~fsspec.core.OpenFile` object or a
        file object, opened in text mode (``'rt'``).
    """
    with utils.Open(file, 'rt') as f:
        yaml_dict = yaml.safe_load(f)

    # template to test yaml_dict items against
    schema = dict(
        im_size=list,
        rpc=dict(
            height_off=float,
            height_scale=float,
            lat_off=float,
            lat_scale=float,
            line_den_coeff=[float],
            line_num_coeff=[float],
            line_off=(int, float),
            line_scale=(int, float),
            long_off=float,
            long_scale=float,
            samp_den_coeff=[float],
            samp_num_coeff=[float],
            samp_off=(int, float),
            samp_scale=(int, float),
        ),
    )

    # validate and convert yaml_dict
    rpc_param_dict = {}
    for filename, rpc_param in yaml_dict.items():
        try:
            if not isinstance(filename, str):
                raise TypeError(f"'{filename}' key is not an an instance of 'str'.")
            utils.validate_collection(schema, rpc_param)
        except (ValueError, TypeError, KeyError) as ex:
            raise ParamError(
                f"'{utils.get_filename(file)}' is not a valid Orthority RPC file: {str(ex)}."
            )
        rpc_param_dict[filename] = dict(
            cam_type=CameraType.rpc, im_size=tuple(rpc_param['im_size']), rpc=rpc_param['rpc']
        )

    return rpc_param_dict


def write_int_param(
    file: str | PathLike | OpenFile | IO[str],
    int_param_dict: dict[str, dict[str, Any]],
    overwrite: bool = False,
) -> None:
    """
    Write interior parameters to an :doc:`Orthority interior parameter file
    <../file_formats/oty_int>`.

    :param file:
        File to write.  Can be a path or URI string, an :class:`~fsspec.core.OpenFile` object or
        a file object, opened in text mode (``'wt'``).
    :param int_param_dict:
        Interior parameters to write.
    :param overwrite:
        Whether to overwrite the file if it exists.
    """
    # convert 'cam_type' key to 'type' & make the converted item the first in the dict
    yaml_dict = {}
    for cam_id, int_param in int_param_dict.items():
        yaml_param = {}
        if 'cam_type' in int_param:
            yaml_param['type'] = int_param['cam_type'].value
        yaml_param.update(**{k: v for k, v in int_param.items() if k != 'cam_type'})
        yaml_dict[cam_id] = yaml_param

    with utils.Open(file, 'wt', overwrite=overwrite) as f:
        yaml.safe_dump(yaml_dict, f, sort_keys=False, indent=4, default_flow_style=None)


def write_ext_param(
    file: str | PathLike | OpenFile | IO[str],
    ext_param_dict: dict[str, dict[str, Any]],
    crs: str | CRS,
    overwrite: bool = False,
) -> None:
    """
    Write exterior parameters to an :doc:`Orthority exterior parameter file
    <../file_formats/oty_ext>`.

    :param file:
        File to write.  Can be a path or URI string, an :class:`~fsspec.core.OpenFile` object or
        a file object, opened in text mode (``'wt'``).
    :param ext_param_dict:
        Exterior parameters to write.
    :param crs:
        CRS of the world coordinate system as an EPSG, WKT or proj4 string; or
        :class:`~rasterio.crs.CRS` object.
    :param overwrite:
        Whether to overwrite the file if it exists.
    """
    feat_list = []
    lla_crs = _default_lla_crs

    try:
        crs = CRS.from_string(crs) if isinstance(crs, str) else crs
    except RioCrsError as ex:
        raise CrsError(f"Could not interpret 'crs': {str(ex)}.")
    if not crs.is_projected:
        raise CrsError(f"'crs' should be a projected system.")

    for src_file, ext_param in ext_param_dict.items():
        xyz = ext_param['xyz']
        lla = transform(crs, lla_crs, [xyz[0]], [xyz[1]], [xyz[2]])
        lla = [lla[0][0], lla[1][0], lla[2][0]]  # (lon, lat) order for GeoJSON point
        props_dict = dict(
            filename=src_file, camera=ext_param['camera'], xyz=xyz, opk=ext_param['opk']
        )
        geom_dict = dict(type='Point', coordinates=list(lla))
        feat_dict = dict(type='Feature', properties=props_dict, geometry=geom_dict)
        feat_list.append(feat_dict)

    json_dict = dict(type='FeatureCollection', world_crs=crs.to_string(), features=feat_list)

    with utils.Open(file, 'wt', overwrite=overwrite) as f:
        json.dump(json_dict, f, indent=4)


def write_rpc_param(
    file: str | PathLike | OpenFile | IO[str],
    rpc_param_dict: dict[str, dict[str, Any]],
    overwrite: bool = False,
) -> None:
    """
    Write RPC parameters to an :doc:`Orthority RPC parameter file <../file_formats/oty_rpc>`.

    :param file:
        File to write.  Can be a path or URI string, an :class:`~fsspec.core.OpenFile` object or
        a file object, opened in text mode (``'wt'``).
    :param rpc_param_dict:
        RPC parameters to write.
    :param overwrite:
        Whether to overwrite the file if it exists.
    """
    # copy rpc_param_dict to yaml_dict without 'cam_type' item(s)
    yaml_dict = {}
    for filename, rpm_param in rpc_param_dict.items():
        yaml_dict[filename] = {k: v for k, v in rpm_param.items() if k != 'cam_type'}

    with utils.Open(file, 'wt', overwrite=overwrite) as f:
        yaml.safe_dump(yaml_dict, f, sort_keys=False, indent=4, default_flow_style=None)


def _rpy_to_rotation(rpy: tuple[float, float, float]) -> np.ndarray:
    """Convert the given (roll, pitch, yaw) angles in radians to a rotation matrix."""
    # see https://s3.amazonaws.com/mics.pix4d.com/KB/documents/Pix4D_Yaw_Pitch_Roll_Omega_to_Phi_Kappa_angles_and_conversion.pdf
    roll, pitch, yaw = rpy
    R_x = np.array([[1, 0, 0], [0, np.cos(roll), -np.sin(roll)], [0, np.sin(roll), np.cos(roll)]])
    R_y = np.array(
        [[np.cos(pitch), 0, np.sin(pitch)], [0, 1, 0], [-np.sin(pitch), 0, np.cos(pitch)]]
    )
    R_z = np.array([[np.cos(yaw), -np.sin(yaw), 0], [np.sin(yaw), np.cos(yaw), 0], [0, 0, 1]])
    return R_z.dot(R_y).dot(R_x)


def _opk_to_rotation(opk: tuple[float, float, float]) -> np.ndarray:
    """Convert the given (omega, phi, kappa) angles in radians to a rotation matrix."""
    # see https://s3.amazonaws.com/mics.pix4d.com/KB/documents/Pix4D_Yaw_Pitch_Roll_Omega_to_Phi_Kappa_angles_and_conversion.pdf
    omega, phi, kappa = opk
    R_x = np.array(
        [[1, 0, 0], [0, np.cos(omega), -np.sin(omega)], [0, np.sin(omega), np.cos(omega)]]
    )
    R_y = np.array([[np.cos(phi), 0, np.sin(phi)], [0, 1, 0], [-np.sin(phi), 0, np.cos(phi)]])
    R_z = np.array(
        [[np.cos(kappa), -np.sin(kappa), 0], [np.sin(kappa), np.cos(kappa), 0], [0, 0, 1]]
    )
    return R_x.dot(R_y).dot(R_z)


def _rotation_to_opk(R: np.ndarray) -> tuple[float, float, float]:
    """Convert the given rotation matrix to the (omega, phi, kappa) angles in radians."""
    # see https://s3.amazonaws.com/mics.pix4d.com/KB/documents/Pix4D_Yaw_Pitch_Roll_Omega_to_Phi_Kappa_angles_and_conversion.pdf
    omega = np.arctan2(-R[1, 2], R[2, 2])
    phi = np.arcsin(R[0, 2])
    kappa = np.arctan2(-R[0, 1], R[0, 0])
    return omega, phi, kappa


def _aa_to_opk(aa: tuple[float, float, float]) -> tuple[float, float, float]:
    """Convert the given angle axis vector (OpenSfM / OpenCV convention) to (omega, phi,
    kappa) angles in radians (PATB convention)."""
    # convert ODM angle/axis to rotation matrix (see
    # https://github.com/mapillary/OpenSfM/issues/121)
    R = cv2.Rodrigues(np.array(aa))[0].T

    # rotate from OpenSfM / OpenCV to PATB convention
    R = R.dot(np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]]))

    # extract OPK from R
    omega, phi, kappa = _rotation_to_opk(R)
    return omega, phi, kappa


def _rpy_to_opk(
    rpy: tuple[float, float, float],
    lla: tuple[float, float, float],
    crs: CRS,
    lla_crs: CRS,
    C_bB: Sequence[float] | np.ndarray | None = None,
) -> tuple[float, float, float]:
    """
    Convert (roll, pitch, yaw) to (omega, phi, kappa) angles for a given CRS.

    Coordinate conventions are as `defined by Pix4D
    <https://s3.amazonaws.com/mics.pix4d.com/KB/documents
    /Pix4D_Yaw_Pitch_Roll_Omega_to_Phi_Kappa_angles_and_conversion.pdf.>`__

    Note this conversion assumes ``crs`` is orthogonal and uniform scale (cartesian) in the
    vicinity of ``lla``.  ``crs`` should be conformal, and have its central meridian(s) close to
    ``lla`` for this assumption to be accurate.

    :param rpy:
        (roll, pitch, yaw) angles to rotate from body to navigation coordinates (radians).
    :param lla:
        (latitude, longitude, altitude) geographic coordinates of the camera (in units of
        ``lla_crs``).
    :param crs:
        CRS of the world coordinate system as an EPSG, WKT or proj4 string; or
        :class:`~rasterio.crs.CRS` object.
    :param lla_crs:
        CRS of the ``lla`` geographic coordinates as an EPSG, WKT or proj4 string; or
        :class:`~rasterio.crs.CRS` object.
    :param C_bB:
        Optional camera to body rotation matrix.  Defaults to:
        ``[[0, 1, 0], [1, 0, 0], [0, 0, -1]]``, which describes typical drone geometry.

    :return:
        (omega, phi, kappa) angles in radians, to rotate from camera (PATB convention) to world
        coordinates.
    """
    # Adapted from the OpenSfM exif module
    # https://github.com/mapillary/OpenSfM/blob/main/opensfm/exif.py and Pix4D doc
    # https://s3.amazonaws.com/mics.pix4d.com/KB/documents/Pix4D_Yaw_Pitch_Roll_Omega_to_Phi_Kappa_angles_and_conversion.pdf.

    # Note that what the Pix4D reference calls Object (E), and Image (B) coordinates, orthority
    # calls world and camera coordinates respectively.

    # RPY rotates from body (b) to navigation (n) coordinates, and OPK from camera ('Image' B) to
    # world ('Object' E) coordinates.  Body (b) coordinates are defined as x->front, y->right,
    # z->down, and navigation (n) coordinates as x->N, y->E, z->down.  Navigation is a tangent
    # plane type system and shares its origin with body (b) at ``lla``. Camera (B) coordinates
    # are in the Pix4D / PATB convention i.e. x->right, y->top, z->back (looking through the
    # camera at the scene), and are aligned with & centered on the body (b) system.  World (E)
    # coordinates are ENU i.e. x->E, y->N, z->up.
    lla = np.array(lla)
    crs = CRS.from_string(crs) if isinstance(crs, str) else crs

    # find rotation matrix C_nb, to rotate from body (b) to navigation (n) coordinates.
    C_nb = _rpy_to_rotation(rpy)

    # find rotation matrix C_En, to rotate from navigation (n) to world (object E) coordinates
    delta = 1e-7
    lla1 = lla + (delta, 0, 0)
    lla2 = lla - (delta, 0, 0)

    # p1 & p2 must be in the world CRS, not ECEF as might be understood from the references
    p1 = np.array(transform(lla_crs, crs, [lla1[1]], [lla1[0]], [lla1[2]])).squeeze()
    p2 = np.array(transform(lla_crs, crs, [lla2[1]], [lla2[0]], [lla2[2]])).squeeze()

    # approximate the relative alignment of world and navigation systems in the vicinity of lla
    x_np = p1 - p2
    m = np.linalg.norm(x_np)
    x_np /= m  # unit vector in navigation system N direction
    z_np = np.array([0, 0, -1]).T
    y_np = np.cross(z_np, x_np)
    C_En = np.array([x_np, y_np, z_np]).T

    # C_bB is the rotation from camera (image B) to body (b) coordinates
    C_bB = np.array(C_bB) if C_bB is not None else np.array([[0, 1, 0], [1, 0, 0], [0, 0, -1]])

    # combine C_En, C_nb, C_bB to find rotation from camera (B) to world (E) coordinates.
    C_EB = C_En.dot(C_nb).dot(C_bB)

    # return OPK angles extracted from C_EB
    omega, phi, kappa = _rotation_to_opk(C_EB)
    return omega, phi, kappa


class FrameReader(ABC):
    """
    Base frame camera parameter reader.

    :param crs:
        CRS of the world coordinate system as an EPSG, WKT or proj4 string; or
        :class:`~rasterio.crs.CRS` object.
    :param lla_crs:
        CRS of input geographic coordinates (if any), as an EPSG, WKT or proj4 string; or
        :class:`~rasterio.crs.CRS` object.
    """

    @abstractmethod
    def __init__(self, crs: str | CRS = None, lla_crs: str | CRS = _default_lla_crs) -> None:
        self._crs, self._lla_crs = self._parse_crss(crs, lla_crs)

    @staticmethod
    def _parse_crss(crs: str | CRS, lla_crs: str | CRS) -> tuple[CRS, CRS]:
        """Validate and convert CRSs."""
        if crs:
            try:
                crs = CRS.from_string(crs) if isinstance(crs, str) else crs
            except RioCrsError as ex:
                raise CrsError(f"Could not interpret 'crs': {str(ex)}.")
            if not crs.is_projected:
                raise CrsError(f"'crs' should be a projected system.")

        if lla_crs:
            try:
                lla_crs = CRS.from_string(lla_crs) if isinstance(lla_crs, str) else lla_crs
            except RioCrsError as ex:
                raise CrsError(f"Could not interpret 'lla_crs': {str(ex)}.")
            if not lla_crs.is_geographic:
                raise CrsError(f"'lla_crs' should be a geographic system.")
        return crs, lla_crs

    @property
    def crs(self) -> CRS:
        """CRS of the world coordinate system."""
        return self._crs

    @abstractmethod
    def read_ext_param(self) -> dict[str, dict[str, Any]]:
        """Read exterior camera parameters."""
        pass


class CsvReader(FrameReader):
    """
    Exterior parameter reader for a CSV file.

    Reads tabular data from a CSV file with a row per source image and column fields for image
    file name, camera position, orientation and ID.

    See the :doc:`CSV documentation <../file_formats/csv>` for details on supported fields and
    formats.

    :param file:
        File to read.  Can be a path or URI string, an :class:`~fsspec.core.OpenFile` object or a
        file object, opened in text mode (``'rt'``).
    :param crs:
        CRS of the world coordinate system as an EPSG, WKT or proj4 string; or
        :class:`~rasterio.crs.CRS` object.  If set to ``None`` (the default), and the file
        contains (``x``, ``y``, ``z``) world coordinate positions, a CRS can be provided via a
        :file:`.prj` sidecar file.  If set to ``None``, and the file contains (``latitude``,
        ``longitude``, ``altitude``) and (``roll``, ``pitch``, ``yaw``) fields, a UTM CRS will be
        auto-determined. If the file contains (``latitude``, ``longitude``, ``altitude``) or
        (``roll``, ``pitch``, ``yaw``) fields (but not both), ``crs`` should be supplied.
    :param lla_crs:
        Geographic CRS associated with any (``latitude``, ``longitude``, ``altitude``) position
        and/or (``roll``, ``pitch``, ``yaw``) values in the file (as an EPSG, WKT or proj4
        string; or :class:`~rasterio.crs.CRS` object).
    :param fieldnames:
        List of names specifying the CSV fields.  If set to ``None`` (the default), names will be
        read from the file header if it exists.  If ``fieldnames`` is supplied, any existing file
        header is ignored.  See the :doc:`CSV documentation <../file_formats/csv>` for recognised
        and required field names.
    :param dialect:
        :class:`~csv.Dialect` object specifying the CSV delimiter, quote character etc. If set to
        ``None`` (the default), this is auto-detected from the file.
    :param radians:
        Whether orientation angles are in radians (``True``), or degrees (``False``).
    """

    _type_schema = dict(
        filename=str,
        x=float,
        y=float,
        z=float,
        latitude=float,
        longitude=float,
        altitude=float,
        omega=float,
        phi=float,
        kappa=float,
        roll=float,
        pitch=float,
        yaw=float,
        camera=lambda x: str(x) if x else x,
    )
    _legacy_fieldnames = ['filename', 'x', 'y', 'z', 'omega', 'phi', 'kappa']

    def __init__(
        self,
        file: str | PathLike | OpenFile | IO[str],
        crs: str | CRS = None,
        lla_crs: str | CRS = _default_lla_crs,
        fieldnames: Sequence[str] = None,
        dialect: Dialect = None,
        radians: bool = False,
        **kwargs,
    ) -> None:
        # TODO: allow other coordinate conventions for opk / rpy (bluh, odm, patb)
        FrameReader.__init__(self, crs, lla_crs=lla_crs)
        self._file = file
        self._radians = radians

        # get / create an OpenFile object to use when finding a .prj file
        self._ofile = None
        if isinstance(file, OpenFile):
            self._ofile = file
        elif isinstance(file, (str, PathLike)):
            self._ofile = fsspec.open(os.fspath(file), 'rt')

        # read file once into a buffer (newline=None works around some delimiter detection problems
        # with newline='')
        with utils.Open(self._ofile or file, 'rt', newline=None) as f:
            self._buffer = StringIO(f.read())

        self._fieldnames, self._dialect, self._has_header, self._format = self._parse_file(
            self._buffer, fieldnames=fieldnames, dialect=dialect
        )
        self._crs = self._crs or self._get_crs()

    @staticmethod
    def _parse_fieldnames(fieldnames: Sequence[str]) -> CsvFormat:
        """Validate a list of header or user field names, and return the corresponding
        :class:`CsvFormat`.
        """
        if 'filename' not in fieldnames:
            raise ParamError(f"Fields should include 'filename'.")

        has_xyz = {'x', 'y', 'z'}.issubset(fieldnames)
        has_lla = {'latitude', 'longitude', 'altitude'}.issubset(fieldnames)
        has_opk = {'omega', 'phi', 'kappa'}.issubset(fieldnames)
        has_rpy = {'roll', 'pitch', 'yaw'}.issubset(fieldnames)

        if not (has_xyz or has_lla):
            raise ParamError(
                f"Fields should include 'x', 'y' & 'z', or 'latitude', 'longitude' & 'altitude'."
            )
        if not (has_opk or has_rpy):
            raise ParamError(
                f"Fields should include 'omega', 'phi' & 'kappa', or 'roll', 'pitch' & 'yaw'."
            )

        # dictionary with key = (has_xyz, has_opk) and value = CsvFormat
        format_dict = {
            (True, True): CsvFormat.xyz_opk,
            (True, False): CsvFormat.xyz_rpy,
            (False, True): CsvFormat.lla_opk,
            (False, False): CsvFormat.lla_rpy,
        }
        return format_dict[(has_xyz, has_opk)]

    @staticmethod
    def _parse_file(
        buffer: StringIO, fieldnames: Sequence[str] = None, dialect: Dialect = None
    ) -> tuple[Sequence[str], Dialect, bool, CsvFormat]:
        """Determine and validate the CSV file format."""

        def strip_lower_strlist(str_list: Sequence[str]) -> list[str]:
            """Strip and lower the case of a string list."""
            return [str_item.strip().lower() for str_item in str_list]

        # read a sample of the csv file
        buffer.seek(0)
        sample = buffer.read(10000)

        # auto-detect dialect if not provided
        sniffer = Sniffer()
        if not dialect:
            dialect = sniffer.sniff(sample, delimiters=',;: \t')
            dialect.skipinitialspace = True

        # find field names and parse to determine format
        has_header = sniffer.has_header(sample)
        if not fieldnames:
            if has_header:  # read field names from header
                fieldnames = next(iter(csv.reader(sample.splitlines(), dialect=dialect)))
            else:  # assume fields in legacy format
                fieldnames = CsvReader._legacy_fieldnames
        fieldnames = strip_lower_strlist(fieldnames)
        csv_fmt = CsvReader._parse_fieldnames(fieldnames)

        return fieldnames, dialect, has_header, csv_fmt

    def _find_lla_rpy_crs(self) -> CRS:
        """Return a UTM CRS covering the mean of the camera positions in a
        :attr:`~CsvFormat.lla_rpy` format file.
        """
        latlons = []
        self._buffer.seek(0)
        reader = DictReader(self._buffer, fieldnames=self._fieldnames, dialect=self._dialect)
        if self._has_header:
            _ = next(iter(reader))
        for row in reader:
            row = {k: self._type_schema[k](v) for k, v in row.items() if k in self._type_schema}
            latlon = (row['latitude'], row['longitude'])
            latlons.append(latlon)

        mean_latlon = np.array(latlons).mean(axis=0)
        return utils.utm_crs_from_latlon(*mean_latlon)

    def _get_crs(self) -> CRS:
        """Read / auto-determine and validate a CRS when no user CRS was supplied."""
        crs = None
        filename = utils.get_filename(self._file)
        if self._format is CsvFormat.xyz_opk or self._format is CsvFormat.xyz_rpy:
            if self._ofile:
                # read CRS of xyz positions / opk orientations from .prj file, if it exists
                prj_path = self._ofile.path[: self._ofile.path.rfind('.')] + '.prj'
                prj_name = Path(prj_path).name
                try:
                    with self._ofile.fs.open(prj_path, 'rt') as f:
                        crs_str = f.read()
                    crs = CRS.from_string(crs_str)

                    if not crs.is_projected:
                        raise ParamError(f"CRS in '{prj_name}' should be a projected system.")

                    logger.debug(f"Using '{prj_name}' CRS: '{crs.to_string()}'")
                except FileNotFoundError as ex:
                    logger.debug(f"Could not open '{prj_name}': {str(ex)}.")
                except RioCrsError as ex:
                    raise ParamError(f"Could not interpret CRS in '{prj_name}': {str(ex)}.")
            else:
                # a file object was passed to __init__ so the CSV file path / URI is unknown and a
                # .prj file cannot be found
                logger.debug(f"Cannot read a .prj file with a CSV file object.")

            if not crs and self._format is CsvFormat.xyz_rpy:
                raise CrsMissingError(f"'crs' should be specified for positions in '{filename}'.")

        elif self._format is CsvFormat.lla_rpy:
            # find a UTM CRS to transform the lla positions & rpy orientations into
            crs = self._find_lla_rpy_crs()
            logger.debug(f"Using auto UTM CRS: '{crs.to_proj4()}'")

        elif self._format is CsvFormat.lla_opk:
            # a user-supplied opk CRS is required to project lla into
            raise CrsMissingError(f"'crs' should be specified for orientations in '{filename}'.")

        return crs

    def _convert(self, row: dict[str, float], radians=False) -> tuple[tuple, tuple]:
        """Convert a CSV row dictionary to (x, y, z) position, and (omega, phi,
        kappa) orientation.
        """
        if self._format.is_xyz:
            xyz = (row['x'], row['y'], row['z'])
        else:
            lla = (row['latitude'], row['longitude'], row['altitude'])
            xyz = transform(self._lla_crs, self._crs, [lla[1]], [lla[0]], [lla[2]])
            xyz = tuple([coord[0] for coord in xyz])

        if self._format.is_opk:
            opk = (row['omega'], row['phi'], row['kappa'])
            opk = opk if radians else tuple(np.radians(opk).tolist())
        else:
            rpy = (row['roll'], row['pitch'], row['yaw'])
            rpy = rpy if radians else tuple(np.radians(rpy).tolist())
            if self._format.is_xyz:
                lla = transform(self._crs, self._lla_crs, [xyz[0]], [xyz[1]], [xyz[2]])
                lla = (lla[1][0], lla[0][0], lla[2][0])  # x, y order -> lat, lon order
            opk = _rpy_to_opk(rpy, lla, self._crs, lla_crs=self._lla_crs)

        return xyz, opk

    def read_ext_param(self) -> dict[str, dict[str, Any]]:
        ext_param_dict = {}
        self._buffer.seek(0)
        reader = DictReader(self._buffer, fieldnames=self._fieldnames, dialect=self._dialect)

        if self._has_header:  # read header
            _ = next(iter(reader))

        for row in reader:
            filename = Path(row['filename']).name
            row = {k: self._type_schema[k](v) for k, v in row.items() if k in self._type_schema}
            xyz, opk = self._convert(row, radians=self._radians)
            ext_param_dict[filename] = dict(xyz=xyz, opk=opk, camera=row.get('camera', None))

        return ext_param_dict


class OsfmReader(FrameReader):
    """
    Interior and exterior parameter reader for an OpenSfM :file:`reconstruction.json` file.

    See the :doc:`format documentation <../file_formats/opensfm>` for supported camera models.

    :param file:
        File to read.  Can be a path or URI string, an :class:`~fsspec.core.OpenFile` object or a
        file object, opened in text mode (``'rt'``).
    :param crs:
        CRS of the world coordinate system as an EPSG, WKT or proj4 string; or
        :class:`~rasterio.crs.CRS` object.  If set to ``None`` (the default), a UTM CRS will be
        auto-determined.
    :param lla_crs:
        CRS of the ``reference_lla`` value in the :file:`reconstruction.json` file as an EPSG, WKT
        or proj4 string; or :class:`~rasterio.crs.CRS` object.
    """

    # TODO: OSfM reconstruction is in a topocentric system, so the transfer of 3D cartesian
    #  positions & rotations into a 2D+1D UTM CRS is an approximation, with similar issues to the
    #  Pix4D RPY->OPK conversion.  Ideally the orthorectification should also happen in this
    #  topocentric system, with the DEM being transformed into it. Then the orthorectified image
    #  can be reprojected to UTM.
    def __init__(
        self,
        file: str | PathLike | OpenFile | IO[str],
        crs: str | CRS = None,
        lla_crs: str | CRS = CRS.from_epsg(4326),
        **kwargs,
    ) -> None:
        FrameReader.__init__(self, crs=crs, lla_crs=lla_crs)
        self._json_dict = self._read_json_dict(file)
        if not self._crs:
            self._crs = self._find_utm_crs()
            logger.debug(f"Using auto UTM CRS: '{self._crs.to_proj4()}'")

    @staticmethod
    def _read_json_dict(file: str | PathLike | OpenFile | IO[str]) -> dict[str, dict]:
        """Read and validate the reconstruction JSON file."""
        with utils.Open(file, 'rt') as f:
            json_data = json.load(f)

        schema = [
            dict(
                cameras=dict,
                shots=dict,
                reference_lla=dict(latitude=float, longitude=float, altitude=float),
            )
        ]
        try:
            utils.validate_collection(schema, json_data)
        except (ValueError, TypeError, KeyError) as ex:
            raise ParamError(
                f"'{utils.get_filename(file)}' is not a valid OpenSfM reconstruction file: "
                f"{str(ex)}."
            )

        # keep root schema keys and delete the rest
        json_dict = {k: json_data[0][k] for k in schema[0].keys()}
        del json_data
        return json_dict

    def _find_utm_crs(self) -> CRS:
        """Return a UTM CRS that covers the reconstruction reference point."""
        ref_lla = self._json_dict['reference_lla']
        ref_lla = (ref_lla['latitude'], ref_lla['longitude'], ref_lla['altitude'])
        return utils.utm_crs_from_latlon(*ref_lla[:2])

    def read_int_param(self) -> dict[str, dict[str, Any]]:
        """Read interior camera parameters."""
        return _read_osfm_int_param(self._json_dict['cameras'])

    def read_ext_param(self) -> dict[str, dict[str, Any]]:
        # transform reference coordinates to the world CRS
        ref_lla = self._json_dict['reference_lla']
        ref_lla = (ref_lla['latitude'], ref_lla['longitude'], ref_lla['altitude'])
        ref_xyz = transform(
            self._lla_crs,
            self._crs,
            [ref_lla[1]],
            [ref_lla[0]],
            [ref_lla[2]],
        )
        ref_xyz = tuple([coord[0] for coord in ref_xyz])

        ext_param_dict = {}
        for filename, shot_dict in self._json_dict['shots'].items():
            # convert  reconstruction 'translation' and 'rotation' to oty exterior params,
            # adapted from ODM: https://github.com/OpenDroneMap/ODM/blob/master/opendm/shots.py
            R = cv2.Rodrigues(np.array(shot_dict['rotation']))[0].T
            delta_xyz = -R.dot(shot_dict['translation'])
            xyz = tuple((ref_xyz + delta_xyz).tolist())
            # rotate camera coords from OpenSfM / OpenCV to PATB convention
            R_ = R.dot(np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]]))
            opk = _rotation_to_opk(R_)
            cam_id = shot_dict['camera']
            cam_id = cam_id[3:] if cam_id.startswith('v2 ') else cam_id
            ext_param_dict[filename] = dict(xyz=xyz, opk=opk, camera=cam_id)

        return ext_param_dict


class ExifReader(FrameReader):
    """
    Interior and exterior parameter reader for image file(s) with EXIF / XMP tags.

    See the :doc:`format documentation <../file_formats/exif_xmp>` for required tags.

    :param files:
        File(s) to read as a tuple of paths or URI strings, :class:`~fsspec.core.OpenFile`
        objects in binary mode (``'rb'``), or dataset readers.
    :param crs:
        CRS of the world coordinate system as an EPSG, WKT or proj4 string; or
        :class:`~rasterio.crs.CRS` object.  If set to ``None`` (the default), a UTM CRS will be
        auto-determined.
    :param lla_crs:
        CRS of geographic camera coordinates in the EXIF / XMP tags as an EPSG, WKT or proj4 string;
        or :class:`~rasterio.crs.CRS` object.
    :param progress:
        Whether to display a progress bar monitoring the portion of files read.  Can be set to a
        dictionary of arguments for a custom `tqdm <https://tqdm.github.io/docs/tqdm/>`_ bar.
    """

    def __init__(
        self,
        files: Sequence[str | PathLike | OpenFile | rio.DatasetReader],
        crs: str | CRS = None,
        lla_crs: str | CRS = _default_lla_crs,
        progress: bool | dict = False,
        **kwargs,
    ) -> None:
        FrameReader.__init__(self, crs, lla_crs)
        files = files if isinstance(files, Iterable) else [files]
        self._exif_dict = self._read_exif(files, progress)

        if not self._crs and len(self._exif_dict) > 0:
            self._crs = self._find_utm_crs()
            logger.debug(f"Using auto UTM CRS: '{self._crs.to_proj4()}'")

    @staticmethod
    def _read_exif(
        files: Sequence[str | PathLike | OpenFile | rio.DatasetReader], progress: bool | dict
    ) -> dict[str, Exif]:
        """Return a dictionary of Exif objects for the given images."""
        # read exif tags in thread pool, populating exif_dict in same order as files
        exif_dict = {}
        with ThreadPoolExecutor() as executor:
            futures = [executor.submit(Exif, file) for file in files]

            # create / set up progress bar
            if progress is True:
                progress = tqdm(futures, **_default_tqdm_kwargs)
            elif progress is False:
                progress = tqdm(futures, disable=True, leave=False)
            else:
                progress = tqdm(futures, **progress)

            for future in progress:
                exif_obj = future.result()
                exif_dict[exif_obj.filename] = exif_obj

        return exif_dict

    def _find_utm_crs(self) -> CRS:
        """Return a UTM CRS that covers the mean of the camera positions."""
        # TODO: use weighted sum as in OpenSfM, then use ExifReader.crs in oty odm, see :
        #  https://github.com/mapillary/OpenSfM/blob/c6b5acef9376a75b87414d900c258ef876a6413a/opensfm/dataset.py#L985
        llas = []
        for e in self._exif_dict.values():
            if not e.lla:
                raise ParamError(f"No latitude, longitude & altitude tags in '{e.filename}'.")
            llas.append(e.lla)

        mean_latlon = np.array(llas)[:, :2].mean(axis=0)
        return utils.utm_crs_from_latlon(*mean_latlon)
        # return CRS.from_proj4(f'+proj=ortho +lat_0={mean_latlon[0]} +lon_0={mean_latlon[1]}'

    def read_int_param(self) -> dict[str, dict[str, Any]]:
        """Read interior camera parameters."""
        int_param_dict = {}
        for filename, exif in self._exif_dict.items():
            int_param = _read_exif_int_param(exif)
            int_param_dict.update(**int_param)
        return int_param_dict

    def read_ext_param(self) -> dict[str, dict[str, Any]]:
        ext_param_dict = {}
        for filename, exif in self._exif_dict.items():
            ext_param_dict.update(
                **_read_exif_ext_param(exif, crs=self._crs, lla_crs=self._lla_crs)
            )
        return ext_param_dict


class OtyReader(FrameReader):
    """
    Exterior parameter reader for an :doc:`Orthority exterior parameter file
    <../file_formats/oty_ext>`.

    :param file:
        File to read.  Can be a path or URI string, an :class:`~fsspec.core.OpenFile` object or a
        file object, opened in text mode (``'rt'``).
    """

    def __init__(self, file: str | PathLike | OpenFile | IO[str], **kwargs) -> None:
        FrameReader.__init__(self)
        self._crs, self._json_dict = self._read_json_dict(file, self._crs)

    @staticmethod
    def _read_json_dict(file: str | PathLike | OpenFile | IO[str], crs: CRS) -> tuple[CRS, dict]:
        """Read and validate the GeoJSON file."""
        with utils.Open(file, 'rt') as f:
            json_dict = json.load(f)

        schema = dict(
            type='FeatureCollection',
            world_crs=str,
            features=[
                dict(
                    type='Feature',
                    properties=dict(filename=str, camera=None, xyz=list, opk=list),
                    geometry=dict(type='Point', coordinates=list),
                )
            ],
        )

        filename = utils.get_filename(file)
        try:
            utils.validate_collection(schema, json_dict)
        except (ValueError, TypeError, KeyError) as ex:
            raise ParamError(
                f"'{filename}' is not a valid GeoJSON exterior parameter file: {str(ex)}."
            )

        if not crs:
            try:
                crs = CRS.from_string(json_dict['world_crs'])
            except RioCrsError as ex:
                raise ParamError(f"Could not interpret CRS in '{filename}': {str(ex)}.")

        return crs, json_dict

    def read_ext_param(self) -> dict[str, dict[str, Any]]:
        ext_param_dict = {}
        for feat_dict in self._json_dict['features']:
            prop_dict = feat_dict['properties']
            filename = prop_dict['filename']
            ext_parm = dict(
                xyz=tuple(prop_dict['xyz']), opk=tuple(prop_dict['opk']), camera=prop_dict['camera']
            )
            ext_param_dict[filename] = ext_parm
        return ext_param_dict



# Copyright The Orthority Contributors.
#
# This file is part of Orthority.
#
# Orthority is free software: you can redistribute it and/or modify it under the terms of the GNU
# Affero General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# Orthority is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Orthority.
# If not, see <https://www.gnu.org/licenses/>.

"""EXIF / XMP image tag decoding and reading."""
from __future__ import annotations

import logging
import warnings
from os import PathLike
from typing import Sequence
from xml.etree import cElementTree as ET

import numpy as np
import rasterio as rio
from fsspec.core import OpenFile

from orthority import utils
from orthority.errors import OrthorityWarning

logger = logging.getLogger(__name__)

_xmp_schemas = dict(
    dji=dict(
        lla_keys=[
            '{http://www.dji.com/drone-dji/1.0/}GpsLatitude',
            '{http://www.dji.com/drone-dji/1.0/}GpsLongtitude',
            '{http://www.dji.com/drone-dji/1.0/}AbsoluteAltitude',
        ],
        rpy_keys=[
            '{http://www.dji.com/drone-dji/1.0/}GimbalRollDegree',
            '{http://www.dji.com/drone-dji/1.0/}GimbalPitchDegree',
            '{http://www.dji.com/drone-dji/1.0/}GimbalYawDegree',
        ],
        dewarp_key='{http://www.dji.com/drone-dji/1.0/}DewarpData',
        rpy_offsets=(0.0, 90.0, 0.0),
        rpy_gains=(1.0, 1.0, 1.0),
    ),
    # the Sensefly / Sony DSC & Pix4D / Parrot Sequoia keys may refer to the RPY of the drone,
    # not camera, but am including for now
    sensefly=dict(
        lla_keys=[],
        rpy_keys=[
            '{http://ns.sensefly.com/Camera/1.0/}Roll',
            '{http://ns.sensefly.com/Camera/1.0/}Pitch',
            '{http://ns.sensefly.com/Camera/1.0/}Yaw',
        ],
        dewarp_key=None,
        rpy_offsets=(0.0, 0.0, 0.0),
        rpy_gains=(1.0, 1.0, 1.0),
    ),
    pix4d=dict(
        lla_keys=[],
        rpy_keys=[
            '{http://pix4d.com/camera/1.0/}Roll',
            '{http://pix4d.com/camera/1.0/}Pitch',
            '{http://pix4d.com/camera/1.0/}Yaw',
        ],
        dewarp_key=None,
        rpy_offsets=(0.0, 0.0, 0.0),
        rpy_gains=(1.0, 1.0, 1.0),
    ),
)
"""
A schema of known XMP keys.

Uses xml namespace qualified keys which are unique, rather than xmltodict type prefix qualified 
keys, which can have different prefixes referring to the same namespace."""


def _xml_to_flat_dict(xmp_str: str) -> dict[str, str]:
    """Return a flat dictionary for the given XML string."""
    etree = ET.fromstring(xmp_str)
    flat_dict = {}

    def traverse_etree(etree: ET) -> None:
        """Traverse the given XML tree, populating flat_dict with xml element (tag, text) and
        attribute (name, value) pairs.
        """
        flat_dict[etree.tag] = etree.text
        if etree.attrib:
            flat_dict.update(**etree.attrib)
        for child in etree.findall("./*"):
            traverse_etree(child)

    traverse_etree(etree)
    return flat_dict


class Exif:
    # Adapted from https://github.com/mapillary/OpenSfM/blob/main/opensfm/exif.py
    """
    EXIF / XMP image tag extractor for camera model related values.

    :param file:
        Image file to read.  Can be a path or URI string, an :class:`~fsspec.core.OpenFile`
        object in binary mode (``'rb'``), or a dataset reader.
    """

    def __init__(self, file: str | PathLike | OpenFile | rio.DatasetReader):
        self._filename = utils.get_filename(file)
        with utils.suppress_no_georef(), rio.Env(GDAL_NUM_THREADS='ALL_CPUS'), utils.OpenRaster(
            file, 'r'
        ) as ds:
            # NB: avoid calling ds.tag_namespaces() which reads more (all?) of the dataset
            # compared to ds.tags() with known ns=
            exif_dict = ds.tags()
            exif_dict = ds.tags(ns='EXIF') if len(exif_dict) == 0 else exif_dict
            self._im_size = ds.shape[::-1]

            xmp_dict = ds.tags(ns='xml:XMP')
            if len(xmp_dict) > 0:
                xmp_str = xmp_dict['xml:XMP'].strip('xml:XMP=')
                xmp_dict = _xml_to_flat_dict(xmp_str)
            else:
                logger.debug(f"'{self._filename}' contains no XMP metadata")

        self._make, self._model, self._serial = self._get_make_model_serial(exif_dict)
        self._tag_im_size = self._get_tag_im_size(exif_dict)
        self._sensor_size = self._get_sensor_size(exif_dict, self._im_size)
        self._focal_len, self._focal_len_35 = self._get_focal_len(exif_dict)
        self._orientation = self._get_orientation(exif_dict)
        self._lla = self._get_xmp_lla(xmp_dict) or self._get_lla(exif_dict)
        self._rpy = self._get_xmp_rpy(xmp_dict)
        self._dewarp = self._get_xmp_dewarp(xmp_dict)

    def __str__(self):
        lla_str = '({:.4f}, {:.4f}, {:.4f})'.format(*self._lla) if self._lla else 'None'
        rpy_str = '({:.4f}, {:.4f}, {:.4f})'.format(*self._rpy) if self._rpy else 'None'
        dewarp_str = ', '.join([f'{p:.4f}' for p in self._dewarp]) if self._dewarp else 'None'
        return (
            f'Image: {self._filename}'
            f'\nCamera: {self._make} {self._model}'
            f'\nActual image size: {self.im_size}'
            f'\nTagged image size: {self._tag_im_size}'
            f'\nFocal length: {self._focal_len}'
            f'\nFocal length (35mm): {self._focal_len_35}'
            f'\nSensor size: {self._sensor_size}'
            f'\nOrientation: {self._orientation}'
            f'\nLatitude, longitude, altitude: {lla_str}'
            f'\nRoll, pitch, yaw: {rpy_str}'
            f'\nDewarp: {dewarp_str}'
        )

    @property
    def filename(self) -> str:
        """Image filename."""
        return self._filename

    @property
    def make(self) -> str | None:
        """Camera make."""
        return self._make

    @property
    def model(self) -> str | None:
        """Camera model."""
        return self._model

    @property
    def serial(self) -> str | None:
        """Camera serial number."""
        return self._serial

    @property
    def im_size(self) -> tuple[int, int] | None:
        """Actual image (width, height) in pixels."""
        return self._im_size

    @property
    def tag_im_size(self) -> tuple[int, int] | None:
        """Tagged image (width, height) in pixels."""
        return self._tag_im_size

    @property
    def sensor_size(self) -> tuple[float, float] | None:
        """Sensor (width, height) in mm."""
        return self._sensor_size

    @property
    def focal_len(self) -> float | None:
        """Focal length in mm."""
        return self._focal_len

    @property
    def focal_len_35(self) -> float | None:
        """35mm equivalent focal length in mm."""
        return self._focal_len_35

    @property
    def orientation(self) -> int | None:
        """Image orientation code (see https://exiftool.org/TagNames/EXIF.html)."""
        return self._orientation

    @property
    def lla(self) -> tuple[float, float, float] | None:
        """(latitude, longitude, altitude) coordinates with latitude and longitude in decimal
        degrees, and altitude in meters.
        """
        return self._lla

    @property
    def rpy(self) -> tuple[float, float, float] | None:
        """(roll, pitch, yaw) camera/gimbal angles in degrees."""
        return self._rpy

    @property
    def dewarp(self) -> Sequence[float] | None:
        """Dewarp parameters."""
        return self._dewarp

    @staticmethod
    def _get_exif_float(exif_dict: dict[str, str], key: str) -> float | tuple[float, ...] | None:
        """Convert numeric EXIF tag to float(s)."""
        if key not in exif_dict:
            return None
        val_list = [
            float(val_str.strip(' (')) for val_str in exif_dict[key].split(')') if len(val_str) > 0
        ]
        return val_list[0] if len(val_list) == 1 else tuple(val_list)

    @staticmethod
    def _get_make_model_serial(
        exif_dict: dict[str, str]
    ) -> tuple[str | None, str | None, str | None]:
        """Return camera make and model string."""
        make_key = 'EXIF_Make'
        model_key = 'EXIF_Model'
        serial_key = 'EXIF_BodySerialNumber'
        make = exif_dict[make_key].lower() if make_key in exif_dict else None
        model = exif_dict[model_key].lower() if model_key in exif_dict else None
        serial = exif_dict[serial_key] if serial_key in exif_dict else None
        return make, model, serial

    @staticmethod
    def _get_tag_im_size(exif_dict: dict[str, str]) -> tuple[int, int] | None:
        """Return the tagged image (width, height) in pixels."""
        width = Exif._get_exif_float(exif_dict, 'EXIF_PixelXDimension')
        height = Exif._get_exif_float(exif_dict, 'EXIF_PixelYDimension')
        return (int(width), int(height)) if width and height else None

    @staticmethod
    def _get_sensor_size(
        exif_dict: dict[str, str], im_size: tuple[int, int] | np.ndarray
    ) -> tuple[float, float] | None:
        """Return the sensor (width, height) in mm."""

        unit_key = 'EXIF_FocalPlaneResolutionUnit'
        xres_key = 'EXIF_FocalPlaneXResolution'
        yres_key = 'EXIF_FocalPlaneYResolution'

        if unit_key not in exif_dict or xres_key not in exif_dict or yres_key not in exif_dict:
            return None

        # find mm per resolution unit
        unit_code = int(exif_dict["EXIF_FocalPlaneResolutionUnit"])
        mm_per_unit_dict = {
            # https://www.sno.phy.queensu.ca/~phil/exiftool/TagNames/EXIF.html
            2: 25.4,  # inches
            3: 10.0,  # cm
            4: 1.0,  # mm
            5: 0.001,  # um
        }
        mm_per_unit = mm_per_unit_dict.get(unit_code, None)
        if not mm_per_unit:
            warnings.warn(
                f'Unknown focal plane resolution unit: {unit_code}', category=OrthorityWarning
            )
            return None

        # return sensor size in mm
        pixels_per_unit = np.array(
            [Exif._get_exif_float(exif_dict, xres_key), Exif._get_exif_float(exif_dict, yres_key)]
        )
        return tuple((mm_per_unit * np.array(im_size) / pixels_per_unit).tolist())

    @staticmethod
    def _get_focal_len(exif_dict: dict[str, str]) -> tuple[float | None, float | None]:
        """Return the actual and 35mm equivalent focal lengths in mm."""
        focal_35 = Exif._get_exif_float(exif_dict, 'EXIF_FocalLengthIn35mmFilm')
        focal = Exif._get_exif_float(exif_dict, 'EXIF_FocalLength')
        return focal, focal_35

    @staticmethod
    def _get_orientation(exif_dict: dict[str, str]) -> int | None:
        ori_key = 'EXIF_Orientation'
        orientation = int(exif_dict[ori_key]) if ori_key in exif_dict else None
        return orientation

    @staticmethod
    def _get_lla(exif_dict: dict[str, str]) -> tuple[float, float, float] | None:
        """Return the (latitude, longitude, altitude) EXIF image location with latitude, longitude
        in decimal degrees, and altitude in meters.
        """
        lat_ref_key = 'EXIF_GPSLatitudeRef'
        lon_ref_key = 'EXIF_GPSLongitudeRef'
        lat_key = 'EXIF_GPSLatitude'
        lon_key = 'EXIF_GPSLongitude'
        if any([key not in exif_dict for key in [lat_ref_key, lon_ref_key, lat_key, lon_key]]):
            return None

        # get latitude, longitude
        def dms_to_decimal(dms: Sequence[float], ref: str):
            """Convert (degrees, minutes, seconds) tuple to decimal degrees, applying reference
            sign.
            """
            sign = 1 if ref in 'NE' else -1
            return ((dms[2] / 60 + dms[1]) / 60 + dms[0]) * sign

        lat = dms_to_decimal(Exif._get_exif_float(exif_dict, lat_key), exif_dict[lat_ref_key])
        lon = dms_to_decimal(Exif._get_exif_float(exif_dict, lon_key), exif_dict[lon_ref_key])

        # get altitude
        alt = Exif._get_exif_float(exif_dict, 'EXIF_GPSAltitude') or 0.0
        alt_ref = int(exif_dict.get('EXIF_GPSAltitudeRef', '0x00'), 0)
        if alt_ref == 1:
            alt *= -1

        return lat, lon, alt

    @staticmethod
    def _get_xmp_lla(xmp_dict: dict[str, str]) -> tuple[float, float, float] | None:
        """Return the XMP (latitude, longitude, altitude) values if all of them exist. ."""
        for schema_name, xmp_schema in _xmp_schemas.items():
            if sum([lla_key in xmp_dict for lla_key in xmp_schema['lla_keys']]) == 3:
                lla = [float(xmp_dict[lla_key]) for lla_key in xmp_schema['lla_keys']]
                return lla[0], lla[1], lla[2]
        return None

    @staticmethod
    def _get_xmp_rpy(xmp_dict: dict[str, str]) -> tuple[float, float, float] | None:
        """Return the camera / gimbal (roll, pitch, yaw) angles in degrees if they exist."""
        for schema_name, xmp_schema in _xmp_schemas.items():
            if sum([rpy_key in xmp_dict for rpy_key in xmp_schema['rpy_keys']]) == 3:
                rpy = np.array([float(xmp_dict[rpy_key]) for rpy_key in xmp_schema['rpy_keys']])
                rpy *= np.array(xmp_schema['rpy_gains'])
                rpy += np.array(xmp_schema['rpy_offsets'])
                return tuple(rpy.tolist())
        return None

    @staticmethod
    def _get_xmp_dewarp(xmp_dict: dict[str, str]) -> Sequence[float] | None:
        """Return the camera dewarp parameters if they exist."""
        for schema_name, xmp_schema in _xmp_schemas.items():
            dewarp_str = xmp_dict.get(xmp_schema['dewarp_key'], None)
            if dewarp_str:
                return tuple([float(ps) for ps in dewarp_str.split(';')[-1].split(',')])
        return None

    def to_dict(self) -> dict[str, object]:
        """Convert to a property dictionary."""
        return {k: getattr(self, k) for k, v in vars(type(self)).items() if isinstance(v, property)}



# Copyright The Orthority Contributors.
#
# This file is part of Orthority.
#
# Orthority is free software: you can redistribute it and/or modify it under the terms of the GNU
# Affero General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# Orthority is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Orthority.
# If not, see <https://www.gnu.org/licenses/>.

__version__ = '0.4.0'



# Copyright The Orthority Contributors.
#
# This file is part of Orthority.
#
# Orthority is free software: you can redistribute it and/or modify it under the terms of the GNU
# Affero General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# Orthority is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Orthority.
# If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from pathlib import Path

import fsspec
import rasterio as rio

from orthority.exif import Exif


def test_odm_image(odm_image_file: Path):
    """Test reading an image with valid EXIF & XMP tags."""
    exif = Exif(odm_image_file)
    assert exif.filename == Path(odm_image_file).name
    for attr in [
        'make',
        'model',
        'serial',
        'focal_len',
        'focal_len_35',
        'im_size',
        'tag_im_size',
        'lla',
        'rpy',
        'dewarp',
        'orientation',
    ]:
        assert getattr(exif, attr) is not None


def test_exif_image(exif_image_file: Path):
    """Test reading an image with EXIF tags including sensor size, and no XMP tags."""
    exif = Exif(exif_image_file)
    assert exif.filename == Path(exif_image_file).name
    for attr in [
        'make',
        'model',
        'serial',
        'focal_len',
        'focal_len_35',
        'im_size',
        'tag_im_size',
        'sensor_size',
        'lla',
        'orientation',
    ]:
        assert getattr(exif, attr) is not None

    assert exif.rpy is None


def test_ngi_image(ngi_image_file: Path):
    """Test reading an image with no EXIF / XMP tags."""
    exif = Exif(ngi_image_file)
    assert exif.filename == Path(ngi_image_file).name
    assert exif.im_size is not None
    for attr in [
        'make',
        'model',
        'serial',
        'focal_len',
        'focal_len_35',
        'tag_im_size',
        'lla',
        'rpy',
        'sensor_size',
        'dewarp',
        'orientation',
    ]:
        assert getattr(exif, attr) is None


def test_dataset(odm_image_file: str):
    """Test reading an image from an open dataset."""
    with rio.open(odm_image_file, 'r') as ds:
        exif = Exif(ds)
        assert not ds.closed
    assert ds.closed
    assert exif.filename == Path(odm_image_file).name
    assert exif.im_size is not None


def test_open_file(odm_image_file: str):
    """Test reading an image from an fsspec OpenFile instance."""
    ofile = fsspec.open(odm_image_file, 'rb')
    exif = Exif(ofile)
    assert exif.filename == Path(odm_image_file).name
    assert exif.im_size is not None



# Copyright The Orthority Contributors.
#
# This file is part of Orthority.
#
# Orthority is free software: you can redistribute it and/or modify it under the terms of the GNU
# Affero General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# Orthority is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Orthority.
# If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
import tracemalloc
from pathlib import Path
from typing import Sequence

import numpy as np
import pytest
import rasterio as rio
from rasterio.enums import MaskFlags
from rasterio.features import shapes
from rasterio.transform import array_bounds
from rasterio.warp import transform_bounds
from rasterio.windows import from_bounds

from orthority import errors, param_io
from orthority.camera import Camera, create_camera, PinholeCamera
from orthority.enums import CameraType, Compress, Interp
from orthority.ortho import Ortho
from orthority.utils import nan_equals
from tests.conftest import _dem_resolution

logger = logging.getLogger(__name__)


def _validate_ortho_files(
    files: Sequence[Path], cc_thresh: float = 0.75, num_ovl_thresh: int = None
):
    """Validate the similarity of overlapping areas in ortho files."""
    cc_array = np.full((len(files),) * 2, fill_value=np.nan)
    num_ovl_thresh = num_ovl_thresh or np.math.factorial(len(files) - 1)

    for i1, file1 in enumerate(files):
        for i2, file2 in enumerate(files[i1 + 1 :]):
            with rio.open(file1, 'r') as im1, rio.open(file2, 'r') as im2:
                # find windows for the overlap area in each image
                ovl_bl = np.array([im1.bounds[:2], im2.bounds[:2]]).max(axis=0)
                ovl_tr = np.array([im1.bounds[2:], im2.bounds[2:]]).min(axis=0)
                if np.any(ovl_bl >= ovl_tr):
                    continue  # no overlap
                ovl_bounds = [*ovl_bl, *ovl_tr]
                win1 = im1.window(*ovl_bounds)
                win2 = im2.window(*ovl_bounds)

                # read overlap area in each image & find common mask
                array1 = im1.read(1, window=win1, masked=True)
                array2 = im2.read(1, window=win2, masked=True)
                mask = ~(array1.mask | array2.mask)

                # test similarity if overlap area > 5%
                im1_mask = im1.dataset_mask().astype('bool', copy=False)
                if (mask.sum() / im1_mask.sum()) > 0.05:
                    cc = np.corrcoef(array1[mask], array2[mask])
                    log_str = (
                        f"Overlap similarity of '{file1.name}' and '{file2.name}': {cc[0, 1]:.4f}"
                    )
                    logger.info(log_str)
                    cc_array[i1, i2] = cc[0, 1]
                    assert cc[0, 1] > cc_thresh, log_str

    # rough test on number of overlapping files
    assert np.sum(~np.isnan(cc_array)) >= num_ovl_thresh


def test_init(
    rgb_byte_src_file: Path, float_utm34n_dem_file: Path, pinhole_camera: Camera, utm34n_crs: str
):
    """Test Ortho initialisation with specified world CRS."""
    ortho = Ortho(rgb_byte_src_file, float_utm34n_dem_file, pinhole_camera, crs=utm34n_crs)
    with rio.open(float_utm34n_dem_file, 'r') as dem_im:
        dem_crs = dem_im.crs

    assert ortho._crs == rio.CRS.from_string(utm34n_crs)
    assert ortho._dem_crs == dem_crs
    assert ortho._dem_array is not None
    assert ortho._dem_transform is not None


def test_init_src_crs(
    rgb_byte_utm34n_src_file: Path, float_utm34n_dem_file: Path, pinhole_camera: Camera
):
    """Test Ortho initialisation with CRS from source file."""
    ortho = Ortho(rgb_byte_utm34n_src_file, float_utm34n_dem_file, pinhole_camera, crs=None)
    with rio.open(rgb_byte_utm34n_src_file, 'r') as src_im:
        src_crs = src_im.crs
    with rio.open(float_utm34n_dem_file, 'r') as dem_im:
        dem_crs = dem_im.crs

    assert ortho._crs == src_crs
    assert ortho._dem_crs == dem_crs
    assert ortho._dem_array is not None
    assert ortho._dem_transform is not None


@pytest.mark.parametrize('dem_band', [1, 2])
def test_init_dem_band(
    rgb_byte_src_file: Path,
    float_utm34n_dem_file: Path,
    pinhole_camera: Camera,
    utm34n_crs: str,
    dem_band: int,
):
    """Test Ortho initialisation with ``dem_band`` reads the correct DEM band."""
    ortho = Ortho(
        rgb_byte_src_file, float_utm34n_dem_file, pinhole_camera, crs=utm34n_crs, dem_band=dem_band
    )
    with rio.open(float_utm34n_dem_file, 'r') as dem_im:
        dem_bounds = array_bounds(*ortho._dem_array.shape, ortho._dem_transform)
        dem_win = dem_im.window(*dem_bounds)
        dem_array = dem_im.read(indexes=dem_band, window=dem_win).astype('float32')
    assert np.all(ortho._dem_array == dem_array)


def test_init_dem_band_error(
    rgb_byte_src_file: Path, float_utm34n_dem_file: Path, pinhole_camera: Camera, utm34n_crs: str
):
    """Test Ortho initialisation with incorrect ``dem_band`` raises an error."""
    with pytest.raises(ValueError) as ex:
        Ortho(rgb_byte_src_file, float_utm34n_dem_file, pinhole_camera, crs=utm34n_crs, dem_band=3)
    assert 'DEM band' in str(ex.value)


def test_init_nocrs_error(
    rgb_byte_src_file: Path, float_utm34n_dem_file: Path, pinhole_camera: Camera
):
    """Test Ortho initialisation without a CRS raises an error."""
    with pytest.raises(errors.CrsMissingError) as ex:
        _ = Ortho(rgb_byte_src_file, float_utm34n_dem_file, pinhole_camera, crs=None)
    assert 'crs' in str(ex.value)


def test_init_dem_coverage_error(
    rgb_byte_src_file: Path, float_utm34n_dem_file: Path, frame_args: dict, utm34n_crs: str
):
    """Test Ortho initialisation without DEM coverage of ortho bounds raises an error."""
    # create a camera positioned away from dem bounds
    camera = PinholeCamera(**frame_args)
    camera.update((0, 0, 0), (0, 0, 0))

    with pytest.raises(ValueError) as ex:
        _ = Ortho(rgb_byte_src_file, float_utm34n_dem_file, camera, crs=utm34n_crs)
    assert 'DEM' in str(ex.value)


def test_init_horizon_fov_error(
    rgb_byte_src_file: Path, float_utm34n_dem_file: Path, frame_args: dict, utm34n_crs: str
):
    """Test Ortho initialisation with a horizontal FOV camera raises an error."""
    # create a camera pointing away from dem bounds
    camera = PinholeCamera(**frame_args)
    camera.update((0, 0, 0), (np.pi / 2, 0, 0))

    with pytest.raises(ValueError) as ex:
        _ = Ortho(rgb_byte_src_file, float_utm34n_dem_file, camera, crs=utm34n_crs)
    assert 'horizon' in str(ex.value)


def test_dem_above_camera_error(
    rgb_byte_src_file: Path, float_utm34n_dem_file: Path, frame_args: dict, utm34n_crs: str
):
    """Test reading the DEM raises an error when it is higher than the (frame) camera."""
    camera = PinholeCamera(**frame_args)

    # move the camera below the DEM
    _xyz = (*frame_args['xyz'][:2], frame_args['xyz'][2] - 1000)
    camera.update(_xyz, frame_args['opk'])

    # init & reproject
    with pytest.raises(ValueError) as ex:
        _ = Ortho(rgb_byte_src_file, float_utm34n_dem_file, camera, crs=utm34n_crs)
    assert 'DEM' in str(ex.value)


@pytest.mark.parametrize(
    'dem_file, crs',
    [
        ('float_utm34n_dem_file', 'utm34n_crs'),
        ('float_utm34n_msl_dem_file', 'utm34n_egm2008_crs'),
        ('float_wgs84_wgs84_dem_file', 'utm34n_egm2008_crs'),
    ],
)
def test_get_init_dem(
    rgb_byte_src_file: Path,
    pinhole_camera: Camera,
    dem_file: str,
    crs: str,
    request: pytest.FixtureRequest,
):
    """Test the bounds of the initial DEM contain the world / ortho boundary at z=min(DEM),
    with different DEM and world / ortho CRSs, some including vertical CRSs.
    """
    dem_file: Path = request.getfixturevalue(dem_file)
    crs: str = request.getfixturevalue(crs)
    dem_interp = Interp.cubic
    resolution = _dem_resolution

    ortho = Ortho(rgb_byte_src_file, dem_file, pinhole_camera, crs=crs)
    dem_array, dem_transform = ortho._reproject_dem(dem_interp, resolution)
    test_bounds = array_bounds(*dem_array.shape, dem_transform)
    min_z = np.nanmin(dem_array)
    xyz = ortho.camera.world_boundary(min_z)
    ref_bounds = *xyz[:2].min(axis=1), *xyz[:2].max(axis=1)

    assert test_bounds[:2] <= ref_bounds[:2]
    assert test_bounds[2:] >= ref_bounds[2:]


@pytest.mark.parametrize(
    # varying rotations starting at ``rotation`` fixture value & keeping full DEM coverage
    'opk_offset',
    [(0, 0, 0), (-15, 10, 0), (-30, 20, 0)],
)
def test_get_gsd_opk(
    rgb_byte_src_file: Path,
    float_utm34n_dem_file: Path,
    frame_args: dict,
    utm34n_crs: str,
    opk_offset: tuple,
    tmp_path: Path,
):
    """Test the GSD as ortho resolution generates approx as many ortho pixels as source pixels
    with different camera orientations.
    """
    _opk = tuple(np.array(frame_args['opk']) + np.radians(opk_offset))
    camera = PinholeCamera(**frame_args)
    camera.update(xyz=frame_args['xyz'], opk=_opk)
    dem_interp = Interp.cubic

    # find the gsd resolution and masked dem
    ortho = Ortho(rgb_byte_src_file, float_utm34n_dem_file, camera, crs=utm34n_crs, dem_band=2)
    resolution = (ortho._get_gsd(),) * 2
    dem_array, dem_transform = ortho._reproject_dem(dem_interp, resolution)
    dem_array_mask, dem_transform_mask = ortho._mask_dem(
        dem_array, dem_transform, dem_interp, crop=False, mask=True
    )
    mask = ~np.isnan(dem_array_mask)

    assert np.array(camera.im_size).prod() == pytest.approx(mask.sum(), rel=0.05)


@pytest.mark.parametrize(
    'dem_file, crs',
    [
        ('float_utm34n_dem_file', 'utm34n_crs'),
        ('float_utm34n_msl_dem_file', 'utm34n_egm2008_crs'),
        ('float_wgs84_wgs84_dem_file', 'utm34n_egm2008_crs'),
    ],
)
def test_get_gsd_vert_crs(
    rgb_byte_src_file: Path,
    pinhole_camera: Camera,
    dem_file: str,
    crs: str,
    request: pytest.FixtureRequest,
):
    """Test the GSD as ortho resolution generates approx as many ortho pixels as source pixels
    with different DEM and world / ortho CRSs, some with vertical CRSs.
    """
    dem_file: Path = request.getfixturevalue(dem_file)
    crs: str = request.getfixturevalue(crs)
    dem_interp = Interp.cubic

    ortho = Ortho(rgb_byte_src_file, dem_file, pinhole_camera, crs=crs, dem_band=2)

    # find the gsd resolution and masked dem
    resolution = (ortho._get_gsd(),) * 2
    dem_array, dem_transform = ortho._reproject_dem(dem_interp, resolution)
    dem_array_mask, dem_transform_mask = ortho._mask_dem(
        dem_array, dem_transform, dem_interp, crop=True, mask=True
    )
    mask = ~np.isnan(dem_array_mask)

    assert np.array(ortho.camera.im_size).prod() == pytest.approx(mask.sum(), rel=0.05)


@pytest.mark.parametrize(
    'interp, resolution',
    [*zip(Interp, [(10, 10)] * len(Interp)), *zip(Interp, [(50, 50)] * len(Interp))],
)
def test_reproject_dem(
    rgb_byte_src_file: Path,
    float_wgs84_wgs84_dem_file: Path,
    pinhole_camera: Camera,
    utm34n_crs: str,
    interp: Interp,
    resolution: tuple[float, float],
):
    """Test DEM is reprojected when it's CRS & resolution is different to the world / ortho CRS &
    ortho resolution.
    """
    ortho = Ortho(
        rgb_byte_src_file, float_wgs84_wgs84_dem_file, pinhole_camera, crs=utm34n_crs, dem_band=2
    )

    # find initial dem bounds
    init_bounds = array_bounds(*ortho._dem_array.shape, ortho._dem_transform)
    init_bounds = np.array(transform_bounds(ortho._dem_crs, ortho._crs, *init_bounds))

    # reproject
    array, transform = ortho._reproject_dem(interp, resolution)
    bounds = np.array(array_bounds(*array.shape, transform))

    # test validity
    assert transform != ortho._dem_transform
    assert array.shape != ortho._dem_array.shape
    assert np.all(np.abs((transform[0], transform[4])) == resolution)
    assert bounds == pytest.approx(init_bounds, abs=2 * max(resolution))
    assert np.nanmean(array) == pytest.approx(np.nanmean(ortho._dem_array), abs=1e-3)


def test_reproject_dem_crs_equal(
    rgb_byte_src_file: Path, float_utm34n_dem_file: Path, pinhole_camera: Camera, utm34n_crs: str
):
    """Test DEM is not reprojected when it's CRS & resolution are the same as the world CRS & ortho
    resolution.
    """
    ortho = Ortho(rgb_byte_src_file, float_utm34n_dem_file, pinhole_camera, crs=utm34n_crs)

    with rio.open(float_utm34n_dem_file, 'r') as dem_im:
        resolution = dem_im.res
    array, transform = ortho._reproject_dem(Interp.cubic, resolution)

    assert transform == ortho._dem_transform
    assert np.all(nan_equals(array, ortho._dem_array))


@pytest.mark.parametrize(
    'dem_file, crs',
    [
        ('float_utm34n_egm2008_dem_file', 'utm34n_egm96_crs'),
        ('float_utm34n_egm96_dem_file', 'utm34n_egm2008_crs'),
        ('float_utm34n_egm96_dem_file', 'utm34n_wgs84_crs'),
    ],
)
def test_reproject_dem_vert_crs_both(
    rgb_byte_src_file: Path,
    dem_file: str,
    pinhole_camera: Camera,
    crs: str,
    request: pytest.FixtureRequest,
):
    """Test DEM reprojection altitude adjustment when both DEM and ortho vertical CRSs are
    specified.
    """
    dem_file: Path = request.getfixturevalue(dem_file)
    crs: str = request.getfixturevalue(crs)

    ortho = Ortho(rgb_byte_src_file, dem_file, pinhole_camera, crs=crs, dem_band=2)
    array, transform = ortho._reproject_dem(Interp.cubic, _dem_resolution)

    # crop array & transform to correspond to ortho._dem_array & ortho._dem_transform (assumes
    # ortho._dem_array lies inside array, which it should)
    dem_win = (
        rio.windows.from_bounds(
            *array_bounds(*ortho._dem_array.shape, ortho._dem_transform), transform=transform
        )
        .round_offsets()
        .round_lengths()
    )
    test_array = array[dem_win.toslices()]
    test_transform = rio.windows.transform(dem_win, transform)

    assert ortho._dem_crs != ortho._crs
    assert test_transform.almost_equals(ortho._dem_transform, precision=1e-6)
    assert test_array.shape == ortho._dem_array.shape

    mask = ~np.isnan(test_array) & ~np.isnan(ortho._dem_array)
    assert test_array[mask] != pytest.approx(ortho._dem_array[mask], abs=0.1)
    assert test_array[mask] == pytest.approx(ortho._dem_array[mask], abs=abs(Ortho._egm_minmax[0]))


@pytest.mark.parametrize(
    'dem_file, crs',
    [('float_utm34n_dem_file', 'utm34n_egm96_crs'), ('float_utm34n_egm96_dem_file', 'utm34n_crs')],
)
@pytest.mark.skipif(rio.get_proj_version() < (9, 1, 1), reason="requires PROJ 9.1.1 or higher")
def test_reproject_dem_vert_crs_one(
    rgb_byte_src_file: Path,
    dem_file: str,
    pinhole_camera: Camera,
    crs: str,
    request: pytest.FixtureRequest,
):
    """Test DEM reprojection does no altitude adjustment when one of DEM and ortho vertical CRSs
    are specified.
    """
    dem_file: Path = request.getfixturevalue(dem_file)
    crs: str = request.getfixturevalue(crs)

    ortho = Ortho(rgb_byte_src_file, dem_file, pinhole_camera, crs=crs, dem_band=2)
    with rio.open(dem_file, 'r') as dem_im:
        resolution = dem_im.res
    array, transform = ortho._reproject_dem(Interp.cubic, resolution)

    # crop array & transform to correspond to ortho._dem_array & ortho._dem_transform (assumes
    # ortho._dem_array lies inside array, which it should)
    dem_win = (
        rio.windows.from_bounds(
            *array_bounds(*ortho._dem_array.shape, ortho._dem_transform), transform=transform
        )
        .round_offsets()
        .round_lengths()
    )
    test_array = array[dem_win.toslices()]
    test_transform = rio.windows.transform(dem_win, transform)

    assert ortho._dem_crs != ortho._crs
    assert test_transform.almost_equals(ortho._dem_transform, precision=1e-6)
    assert test_array.shape == ortho._dem_array.shape

    mask = ~np.isnan(test_array) & ~np.isnan(ortho._dem_array)
    # prior proj versions promote 2D->3D with ellipsoidal height
    assert test_array[mask] == pytest.approx(ortho._dem_array[mask], abs=1e-3)


def test_reproject_dem_vert_crs_scale(
    rgb_byte_src_file: Path,
    float_utm34n_msl_dem_file: Path,
    pinhole_camera: Camera,
    utm34n_egm2008_crs: str,
):
    """Test DEM reprojection z scaling when DEM height is in feet."""
    ortho = Ortho(
        rgb_byte_src_file,
        float_utm34n_msl_dem_file,
        pinhole_camera,
        crs=utm34n_egm2008_crs,
        dem_band=2,
    )
    array, transform = ortho._reproject_dem(Interp.cubic, _dem_resolution)
    assert np.nanmean(array) == pytest.approx(np.nanmean(ortho._dem_array) / 3.28084, abs=1e-3)


def test_reproject_resolution_error(rgb_pinhole_utm34n_ortho: Ortho):
    """Test DEM reprojection raises an error when the resolution exceeds the ortho bounds."""
    with pytest.raises(ValueError) as ex:
        _, _ = rgb_pinhole_utm34n_ortho._reproject_dem(Interp.cubic, (1000, 1000))
    assert 'resolution' in str(ex.value)


def test_mask_dem(rgb_pinhole_utm34n_ortho: Ortho, tmp_path: Path):
    """Test the similarity of the masked DEM (ortho boundary) and ortho valid data mask (without
    cropping).
    """
    # Notes:
    # - The DEM intersection algorithm is tested more rigorously in
    # test_camera.test_world_boundary_zsurf().
    # - This test should use the pinhole camera model to ensure no artefacts outside the ortho
    # boundary, and DEM < camera height to ensure no ortho artefacts in DEM > camera height
    # areas.  While the DEM mask excludes (boundary) occluded pixels, the ortho image mask does
    # not i.e. to compare these masks, there should be no DEM - ortho occlusion.
    resolution = (3, 3)
    num_pts = 400
    dem_interp = Interp.cubic
    ortho = rgb_pinhole_utm34n_ortho

    # create an ortho image & mask without DEM masking
    dem_array, dem_transform = ortho._reproject_dem(dem_interp, resolution)
    j, i = np.meshgrid(range(0, dem_array.shape[1]), range(0, dem_array.shape[0]), indexing='xy')
    x, y = (dem_transform * rio.Affine.translation(0.5, 0.5)) * [j, i]
    im_array = np.ones((1, *ortho.camera.im_size[::-1]))
    ortho_array, ortho_mask = ortho.camera.remap(
        im_array, x, y, dem_array, nodata=float('nan'), interp=dem_interp
    )
    ortho_mask = ~ortho_mask

    # create the dem mask
    dem_array_mask, dem_transform_mask = ortho._mask_dem(
        dem_array.copy(),
        dem_transform,
        dem_interp,
        crop=False,
        mask=True,
        num_pts=num_pts,
    )
    dem_mask = ~np.isnan(dem_array_mask)

    # test dem mask contains, and is similar to the ortho mask
    assert dem_transform_mask == dem_transform
    assert dem_mask.shape == ortho_mask.shape
    assert dem_mask[ortho_mask].sum() / ortho_mask.sum() > 0.95
    cc = np.corrcoef(dem_mask.flatten(), ortho_mask.flatten())
    assert cc[0, 1] > 0.9

    if False:
        # debug plotting code
        # %matplotlib
        from matplotlib import pyplot
        from rasterio.plot import show

        def plot_poly(mask: np.ndarray, transform=dem_transform, ico='k'):
            """Plot polygons from mask."""
            poly_list = [poly for poly, _ in shapes(mask.astype('uint8'), transform=transform)]

            for poly in poly_list[:-1]:
                coords = np.array(poly['coordinates'][0]).T
                pyplot.plot(coords[0], coords[1], ico)

        for image in (ortho_array, dem_array):
            pyplot.figure()
            show(image, transform=dem_transform, cmap='gray')
            plot_poly(ortho_mask, transform=dem_transform, ico='y--')
            plot_poly(dem_mask, transform=dem_transform, ico='r:')
            pyplot.plot(*ortho._camera._T[:2], 'cx')


def test_mask_dem_crop(rgb_pinhole_utm34n_ortho: Ortho, tmp_path: Path):
    """Test the DEM mask is cropped to mask boundaries."""
    ortho = rgb_pinhole_utm34n_ortho
    resolution = (5, 5)
    num_pts = 400
    dem_interp = Interp.cubic

    # mask the dem without cropping
    dem_array, dem_transform = ortho._reproject_dem(dem_interp, resolution)
    dem_array_mask, dem_transform_mask = ortho._mask_dem(
        dem_array.copy(),
        dem_transform,
        dem_interp,
        crop=False,
        mask=True,
        num_pts=num_pts,
    )
    mask = ~np.isnan(dem_array_mask)

    # crop & mask the dem
    dem_array_crop, dem_transform_crop = ortho._mask_dem(
        dem_array.copy(),
        dem_transform,
        dem_interp,
        crop=True,
        mask=True,
        num_pts=num_pts,
    )
    mask_crop = ~np.isnan(dem_array_crop)

    # find the window of mask_crop in mask
    bounds_crop = array_bounds(*dem_array_crop.shape, dem_transform_crop)
    win_crop = from_bounds(*bounds_crop, dem_transform_mask)

    # sanity testing
    assert dem_transform_mask == dem_transform
    assert np.any(mask_crop)

    # test windowed portion of mask is identical to mask_crop, and unwindowed portion contains no
    # masked pixels
    assert np.all(mask_crop == mask[win_crop.toslices()])
    assert mask.sum() == mask[win_crop.toslices()].sum()

    # test mask_crop extends to the boundaries
    ij = np.where(mask_crop)
    assert np.min(ij, axis=1) == pytest.approx((0, 0), abs=1)
    assert np.max(ij, axis=1) == pytest.approx(np.array(mask_crop.shape) - 1, abs=1)


def test_mask_dem_partial(
    rgb_byte_src_file: Path, float_utm34n_partial_dem_file: Path, frame_args: dict, utm34n_crs: str
):
    """Test the DEM mask excludes DEM nodata and is cropped to mask boundaries."""
    camera = PinholeCamera(**frame_args, distort=True)
    resolution = (5, 5)
    num_pts = 400
    dem_interp = Interp.cubic
    ortho = Ortho(rgb_byte_src_file, float_utm34n_partial_dem_file, camera, utm34n_crs)

    # mask the dem without cropping
    dem_array, dem_transform = ortho._reproject_dem(dem_interp, resolution)
    valid_mask = ~np.isnan(dem_array)
    dem_array_mask, dem_transform_mask = ortho._mask_dem(
        dem_array.copy(),
        dem_transform,
        dem_interp,
        crop=False,
        mask=True,
        num_pts=num_pts,
    )

    # crop & mask the dem
    dem_array_crop, dem_transform_crop = ortho._mask_dem(
        dem_array.copy(),
        dem_transform,
        dem_interp,
        crop=True,
        mask=True,
        num_pts=num_pts,
    )
    mask_crop = ~np.isnan(dem_array_crop)

    # find the window of dem_array_crop in dem_array_mask
    bounds_crop = array_bounds(*dem_array_crop.shape, dem_transform_crop)
    win_crop = from_bounds(*bounds_crop, dem_transform_mask)

    # test the dem contains nodata
    assert not np.all(valid_mask[win_crop.toslices()])

    # test mask_crop extends to the boundaries
    ij = np.where(mask_crop)
    assert np.min(ij, axis=1) == pytest.approx((0, 0), abs=1)
    assert np.max(ij, axis=1) == pytest.approx(np.array(mask_crop.shape) - 1, abs=1)

    # test mask_crop excludes dem nodata
    assert np.all(mask_crop == mask_crop & valid_mask[win_crop.toslices()])


def test_mask_dem_coverage_error(
    rgb_byte_src_file: Path, float_utm34n_dem_file: Path, frame_args: dict, utm34n_crs: str
):
    """Test DEM masking without DEM coverage raises an error."""
    camera = PinholeCamera(**frame_args)

    # init & reproject with coverage
    ortho = Ortho(rgb_byte_src_file, float_utm34n_dem_file, camera, crs=utm34n_crs)
    dem_array, dem_transform = ortho._reproject_dem(Interp.cubic, (30.0, 30.0))

    # update camera for no coverage
    camera.update((0.0, 0.0, 1000.0), (0.0, 0.0, 0.0))

    # test
    with pytest.raises(ValueError) as ex:
        ortho._mask_dem(dem_array, dem_transform, Interp.cubic)
    assert 'boundary' in str(ex.value)


def test_create_ortho_profile_12bit_jpeg(rgb_pinhole_utm34n_ortho: Ortho):
    """Test _create_ortho_profile correctly configures a 12bit jpeg ortho profile."""
    # Note: depending on how rasterio is built, it may or may not support reading/writing 12 bit
    # jpeg compression.  This test just checks the ortho profile is correct.
    with rio.open(rgb_pinhole_utm34n_ortho._src_file, 'r') as src_im:
        ortho_profile, write_mask = rgb_pinhole_utm34n_ortho._create_ortho_profile(
            src_im,
            (1, 1),
            rio.Affine.identity(),
            dtype='uint16',
            compress=Compress.jpeg,
            write_mask=None,
        )

    assert write_mask
    assert ortho_profile['dtype'] == 'uint16'
    assert ortho_profile['compress'] == 'jpeg'
    assert 'nbits' in ortho_profile and ortho_profile['nbits'] == 12


@pytest.mark.parametrize('resolution', [(30.0, 30.0), (60.0, 60.0), (60.0, 30.0)])
def test_process_resolution(rgb_pinhole_utm34n_ortho: Ortho, resolution: tuple, tmp_path: Path):
    """Test ortho ``resolution`` is set correctly."""
    ortho_file = tmp_path.joinpath('test_ortho.tif')
    rgb_pinhole_utm34n_ortho.process(ortho_file, resolution)
    assert ortho_file.exists()

    with rio.open(ortho_file, 'r') as ortho_im:
        assert ortho_im.res == resolution


@pytest.mark.parametrize('interp', [Interp.average, Interp.bilinear, Interp.cubic, Interp.lanczos])
def test_process_interp(rgb_pinhole_utm34n_ortho: Ortho, interp: Interp, tmp_path: Path):
    """Test the process ``interp`` setting by comparing with an ``interp='nearest'`` reference
    ortho.
    """
    resolution = (10, 10)

    ortho_ref_file = tmp_path.joinpath('ref_ortho.tif')
    rgb_pinhole_utm34n_ortho.process(
        ortho_ref_file, resolution, interp=Interp.nearest, compress=Compress.deflate
    )

    ortho_test_file = tmp_path.joinpath('test_ortho.tif')
    rgb_pinhole_utm34n_ortho.process(
        ortho_test_file, resolution, interp=interp, compress=Compress.deflate
    )

    assert ortho_ref_file.exists() and ortho_test_file.exists()
    with rio.open(ortho_ref_file, 'r') as ref_im, rio.open(ortho_test_file, 'r') as test_im:
        ref_array = ref_im.read(masked=True)
        test_array = test_im.read(masked=True)
        assert test_array.shape == ref_array.shape
        assert test_array.mask.sum() == pytest.approx(ref_array.mask.sum(), rel=0.05)
        assert len(np.unique(test_array.compressed())) > len(np.unique(ref_array.compressed()))
        test_array.mask |= ref_array.mask
        ref_array.mask |= test_array.mask
        cc = np.corrcoef(test_array.flatten(), ref_array.flatten())
        assert cc[0, 1] > 0.9
        assert cc[0, 1] != 1.0


@pytest.mark.parametrize('dem_interp', [Interp.bilinear, Interp.cubic, Interp.lanczos])
def test_process_dem_interp(rgb_pinhole_utm34n_ortho: Ortho, dem_interp: Interp, tmp_path: Path):
    """Test the process ``dem_interp`` setting by comparing with an ``dem_interp='nearest'``
    reference ortho.
    """
    # note that Interp.average is skipped as it gives similar upsampling results to Interp.nearest
    resolution = (10, 10)
    ortho_ref_file = tmp_path.joinpath('ref_ortho.tif')
    rgb_pinhole_utm34n_ortho.process(
        ortho_ref_file, resolution, dem_interp=Interp.nearest, compress=Compress.deflate
    )

    ortho_test_file = tmp_path.joinpath('test_ortho.tif')
    rgb_pinhole_utm34n_ortho.process(
        ortho_test_file, resolution, dem_interp=dem_interp, compress=Compress.deflate
    )

    assert ortho_ref_file.exists() and ortho_test_file.exists()
    with rio.open(ortho_ref_file, 'r') as ref_im, rio.open(ortho_test_file, 'r') as test_im:
        ref_array = ref_im.read(masked=True)
        test_win = test_im.window(*ref_im.bounds)
        test_array = test_im.read(masked=True, window=test_win, boundless=True)
        assert test_array.mask.sum() == pytest.approx(ref_array.mask.sum(), rel=0.1)
        test_array.mask |= ref_array.mask
        ref_array.mask |= test_array.mask
        cc = np.corrcoef(test_array.flatten(), ref_array.flatten())
        assert cc[0, 1] > 0.9
        assert cc[0, 1] != 1.0


def test_process_per_band(rgb_pinhole_utm34n_ortho: Ortho, tmp_path: Path):
    """Test ortho equivalence for ``per_band=True/False`` and that ``per_band=True`` uses more
    memory than ``per_band=False``."""
    resolution = (5, 5)

    tracemalloc.start()
    try:
        # create a ref (per_band=True) and test (per_band=False) ortho
        ortho_ref_file = tmp_path.joinpath('ref_ortho.tif')
        rgb_pinhole_utm34n_ortho.process(
            ortho_ref_file, resolution, per_band=True, compress=Compress.deflate
        )
        _, ref_peak = tracemalloc.get_traced_memory()
        tracemalloc.clear_traces()

        ortho_test_file = tmp_path.joinpath('test_ortho.tif')
        rgb_pinhole_utm34n_ortho.process(
            ortho_test_file, resolution, per_band=False, compress=Compress.deflate
        )
        _, test_peak = tracemalloc.get_traced_memory()
    finally:
        tracemalloc.stop()

    # compare memory usage
    assert ref_peak > test_peak

    # compare ref and test orthos
    assert ortho_ref_file.exists() and ortho_test_file.exists()
    with rio.open(ortho_ref_file, 'r') as ref_im, rio.open(ortho_test_file, 'r') as test_im:
        ref_array = ref_im.read()
        test_array = test_im.read()

        assert test_array.shape[0] == 3
        assert test_array.shape == ref_array.shape
        assert np.all(test_array == ref_array)


@pytest.mark.parametrize(
    'camera, camera_und',
    [
        ('pinhole_camera', 'pinhole_camera_und'),
        ('brown_camera', 'brown_camera_und'),
        ('opencv_camera', 'opencv_camera_und'),
        ('fisheye_camera', 'fisheye_camera_und'),
    ],
)
def test_process_distort(
    rgb_byte_src_file: Path,
    float_utm34n_dem_file: Path,
    camera: str,
    camera_und: str,
    utm34n_crs,
    tmp_path: Path,
    request: pytest.FixtureRequest,
):
    """Test ortho similarity for frame cameras with ``distort=True/False`` and ``alpha=1``."""
    # note that these tests are mostly duplicated by test_camera.test_frame_remap_equiv

    camera: Camera = request.getfixturevalue(camera)
    camera_und: Camera = request.getfixturevalue(camera_und)
    ortho = Ortho(rgb_byte_src_file, float_utm34n_dem_file, camera, utm34n_crs)
    ortho_und = Ortho(rgb_byte_src_file, float_utm34n_dem_file, camera_und, utm34n_crs)
    resolution = (3, 3)

    # Create ref (distort=True) and test (distort=False) orthos. Note that
    # distort=False erodes the ortho mask to remove nodata blur so the reference is expected
    # to contain the test mask.
    ortho_ref_file = tmp_path.joinpath('ref_ortho.tif')
    ortho.process(ortho_ref_file, resolution, compress=Compress.deflate)
    ortho_test_file = tmp_path.joinpath('test_ortho.tif')
    ortho_und.process(ortho_test_file, resolution, compress=Compress.deflate)

    # compare ref & test ortho extents, masks and pixels
    assert ortho_ref_file.exists() and ortho_test_file.exists()
    with rio.open(ortho_ref_file, 'r') as ref_im, rio.open(ortho_test_file, 'r') as test_im:
        ref_win = ref_im.window(*test_im.bounds)
        ref_array = ref_im.read(window=ref_win)
        ref_mask = ref_im.dataset_mask(window=ref_win).astype('bool')
        ref_bounds = np.array(ref_im.bounds)
        test_array = test_im.read()
        test_mask = test_im.dataset_mask().astype('bool')
        test_bounds = np.array(test_im.bounds)

        # test ref_mask contains and is similar to test_mask
        assert test_array.shape == ref_array.shape
        assert ref_bounds == pytest.approx(test_bounds, abs=resolution[0])
        assert ref_mask[test_mask].sum() / test_mask.sum() > 0.99
        cc = np.corrcoef(ref_mask.flatten(), test_mask.flatten())
        assert cc[0, 1] > 0.9
        assert (
            cc[0, 1] == pytest.approx(1.0, abs=1e-3)
            if type(camera) == PinholeCamera
            else cc[0, 1] < 1.0
        )

        # test similarity of ortho pixels in intersecting ref-test area
        mask = ref_mask & test_mask
        cc = np.corrcoef(ref_array[:, mask].flatten(), test_array[:, mask].flatten())
        assert cc[0, 1] > 0.95
        assert (
            cc[0, 1] == pytest.approx(1.0, abs=1e-3)
            if type(camera) == PinholeCamera
            else cc[0, 1] < 1.0
        )


@pytest.mark.parametrize(
    'cam_type, dist_param',
    [
        (CameraType.pinhole, {}),
        (CameraType.brown, 'brown_dist_param'),
        (CameraType.opencv, 'opencv_dist_param'),
        (CameraType.fisheye, 'fisheye_dist_param'),
    ],
)
def test_process_alpha(
    cam_type: CameraType,
    dist_param: str,
    frame_args: dict,
    rgb_byte_src_file: Path,
    float_utm34n_dem_file: Path,
    utm34n_crs: str,
    tmp_path: Path,
    request: pytest.FixtureRequest,
):
    """Test ortho with ``alpha=1`` contains and is similar to ortho with ``alpha=0``."""
    # note that these tests are mostly duplicated by test_camera.test_frame_remap_alpha,
    # but without bounds testing
    dist_param: dict = request.getfixturevalue(dist_param) if dist_param else {}
    camera_alpha1 = create_camera(cam_type, **frame_args, **dist_param, alpha=1.0, distort=False)
    camera_alpha0 = create_camera(cam_type, **frame_args, **dist_param, alpha=0.0, distort=False)
    resolution = (3, 3)

    # create a ref (alpha=1) and test (alpha=0) orthos
    ortho = Ortho(rgb_byte_src_file, float_utm34n_dem_file, camera_alpha1, utm34n_crs, dem_band=1)
    ortho_ref_file = tmp_path.joinpath('ref_ortho.tif')
    ortho.process(ortho_ref_file, resolution, compress=Compress.deflate)

    ortho = Ortho(rgb_byte_src_file, float_utm34n_dem_file, camera_alpha0, utm34n_crs, dem_band=1)
    ortho_test_file = tmp_path.joinpath('test_ortho.tif')
    ortho.process(ortho_test_file, resolution, compress=Compress.deflate)

    # compare ref & test ortho extents, masks and pixels
    assert ortho_ref_file.exists() and ortho_test_file.exists()
    with rio.open(ortho_ref_file, 'r') as ref_im, rio.open(ortho_test_file, 'r') as test_im:
        ref_win = ref_im.window(*test_im.bounds)
        ref_array = ref_im.read()
        ref_mask = ref_im.dataset_mask().astype('bool')
        ref_bounds = np.array(ref_im.bounds)
        test_array = test_im.read()
        test_mask = test_im.dataset_mask().astype('bool')
        test_bounds = np.array(test_im.bounds)

        # test ref_mask contains test_mask
        assert test_mask.shape == (ref_win.height, ref_win.width)
        if cam_type is CameraType.pinhole:
            assert np.all(ref_bounds[:2] <= test_bounds[:2]) and np.all(
                ref_bounds[-2:] >= test_bounds[:2]
            )
            assert ref_mask.sum() == test_mask.sum()
        else:
            assert np.all(ref_bounds[:2] < test_bounds[:2]) and np.all(
                ref_bounds[-2:] > test_bounds[:2]
            )
            assert ref_mask.sum() > test_mask.sum()
        ref_mask = ref_mask[ref_win.toslices()]
        assert (ref_mask[test_mask].sum() / test_mask.sum()) > 0.99

        # test similarity of ortho pixels in intersecting ref-test area
        mask = ref_mask & test_mask
        ref_array = ref_array[(slice(0, ref_im.count), *ref_win.toslices())]
        cc = np.corrcoef(ref_array[:, mask].flatten(), test_array[:, mask].flatten())
        assert cc[0, 1] > 0.95


@pytest.mark.parametrize(
    'src_file, write_mask, per_band',
    [
        ('rgb_byte_src_file', True, True),
        ('rgb_byte_src_file', False, True),
        ('rgb_byte_src_file', True, False),
        ('rgb_byte_src_file', False, False),
        ('float_src_file', True, False),
        ('float_src_file', False, False),
    ],
)
def test_process_write_mask_per_band(
    src_file: str,
    float_utm34n_dem_file: Path,
    pinhole_camera: Camera,
    utm34n_crs: str,
    write_mask: bool,
    per_band: bool,
    tmp_path: Path,
    request: pytest.FixtureRequest,
):
    """Test ``write_mask=True`` writes an internal ortho mask irrespective of the value of
    ``per_band``.
    """
    src_file: Path = request.getfixturevalue(src_file)
    ortho = Ortho(src_file, float_utm34n_dem_file, pinhole_camera, utm34n_crs)
    ortho_file = tmp_path.joinpath('test_ortho.tif')
    ortho.process(ortho_file, _dem_resolution, write_mask=write_mask, per_band=per_band)
    assert ortho_file.exists()

    with rio.open(ortho_file, 'r') as ortho_im:
        mask_flag = MaskFlags.per_dataset if write_mask else MaskFlags.nodata
        assert all([mf[0] == mask_flag for mf in ortho_im.mask_flag_enums])


@pytest.mark.parametrize('compress', [None, *Compress])
def test_process_write_mask_compress(
    rgb_pinhole_utm34n_ortho: Ortho, compress: Compress, tmp_path: Path
):
    """Test ``write_mask=None`` writes an internal ortho mask when jpeg compression is used."""
    ortho_file = tmp_path.joinpath('test_ortho.tif')
    rgb_pinhole_utm34n_ortho.process(
        ortho_file, _dem_resolution, write_mask=None, compress=compress
    )
    assert ortho_file.exists()

    with rio.open(ortho_file, 'r') as ortho_im:
        mask_flag = (
            MaskFlags.per_dataset if ortho_im.profile['compress'] == 'jpeg' else MaskFlags.nodata
        )
        assert all([mf[0] == mask_flag for mf in ortho_im.mask_flag_enums])


@pytest.mark.parametrize(
    'src_file, dtype',
    [
        ('rgb_byte_src_file', None),
        ('float_src_file', None),
        # all opencv supported dtypes
        ('rgb_byte_src_file', 'uint8'),
        ('rgb_byte_src_file', 'uint16'),
        ('rgb_byte_src_file', 'int16'),
        ('rgb_byte_src_file', 'float32'),
        ('rgb_byte_src_file', 'float64'),
    ],
)
def test_process_dtype(
    src_file: str,
    float_utm34n_dem_file: Path,
    pinhole_camera: Camera,
    utm34n_crs: str,
    dtype: str,
    tmp_path: Path,
    request: pytest.FixtureRequest,
):
    """Test the ortho ``dtype`` is set correctly."""
    src_file: Path = request.getfixturevalue(src_file)
    ortho = Ortho(src_file, float_utm34n_dem_file, pinhole_camera, utm34n_crs, dem_band=1)
    ortho_file = tmp_path.joinpath('test_ortho.tif')
    ortho.process(ortho_file, _dem_resolution, dtype=dtype)

    assert ortho_file.exists()
    with rio.open(src_file, 'r') as src_im, rio.open(ortho_file, 'r') as ortho_im:
        dtype = src_im.profile['dtype'] if dtype is None else dtype
        assert ortho_im.profile['dtype'] == dtype


@pytest.mark.parametrize('dtype', ['int8', 'uint32', 'int32', 'uint64', 'int64'])
def test_process_dtype_error(rgb_pinhole_utm34n_ortho: Ortho, dtype: str, tmp_path: Path):
    """Test unsupported dtypes raise an error."""
    ortho_file = tmp_path.joinpath('test_ortho.tif')

    with pytest.raises(ValueError) as ex:
        rgb_pinhole_utm34n_ortho.process(ortho_file, _dem_resolution, dtype=dtype)
    assert dtype in str(ex.value)


@pytest.mark.parametrize(
    # all opencv supported dtypes
    'dtype',
    ['uint8', 'uint16', 'int16', 'float32', 'float64'],
)
def test_process_nodata(rgb_pinhole_utm34n_ortho: Ortho, dtype: str, tmp_path: Path):
    """Test the ortho ``nodata`` is set correctly."""
    ortho_file = tmp_path.joinpath('test_ortho.tif')
    rgb_pinhole_utm34n_ortho.process(
        ortho_file, _dem_resolution, dtype=dtype, compress=Compress.deflate
    )

    assert ortho_file.exists()
    with rio.open(ortho_file, 'r') as ortho_im:
        assert ortho_im.profile['dtype'] in Ortho._nodata_vals
        assert nan_equals(ortho_im.profile['nodata'], Ortho._nodata_vals[ortho_im.profile['dtype']])


@pytest.mark.parametrize(
    'src_file, compress',
    [
        ('rgb_byte_src_file', None),
        ('rgb_byte_src_file', Compress.jpeg),
        ('rgb_byte_src_file', Compress.deflate),
        ('float_src_file', None),
        ('float_src_file', Compress.deflate),
    ],
)
def test_process_compress(
    src_file: str,
    float_utm34n_dem_file: Path,
    pinhole_camera: Camera,
    utm34n_crs: str,
    compress: Compress,
    tmp_path: Path,
    request: pytest.FixtureRequest,
):
    """Test the ortho compression, interleaving and photometric interpretation are set correctly."""
    src_file: Path = request.getfixturevalue(src_file)
    ortho = Ortho(src_file, float_utm34n_dem_file, pinhole_camera, utm34n_crs, dem_band=1)
    ortho_file = tmp_path.joinpath('test_ortho.tif')
    ortho.process(ortho_file, _dem_resolution, compress=compress)

    assert ortho_file.exists()
    with rio.open(src_file, 'r') as src_im, rio.open(ortho_file, 'r') as ortho_im:
        if compress is None:
            compress = Compress.jpeg if src_im.dtypes[0] == 'uint8' else Compress.deflate
        interleave, photometric = (
            ('pixel', 'ycbcr')
            if compress == Compress.jpeg and src_im.count == 3
            else ('band', 'minisblack')
        )
        assert ortho_im.profile['compress'] == compress.name
        assert ortho_im.profile['interleave'] == interleave
        if 'photometric' in ortho_im.profile:
            assert ortho_im.profile['photometric'] == photometric


def test_process_compress_jpeg_error(
    float_src_file: Path,
    float_utm34n_dem_file: Path,
    pinhole_camera: Camera,
    utm34n_crs: str,
    tmp_path: Path,
):
    """Test that jpeg compression raises an error when the source image dtype is not uint8 or
    uint16.
    """
    ortho = Ortho(float_src_file, float_utm34n_dem_file, pinhole_camera, utm34n_crs, dem_band=1)
    ortho_file = tmp_path.joinpath('test_ortho.tif')

    with pytest.raises(ValueError) as ex:
        ortho.process(ortho_file, (5, 5), compress=Compress.jpeg)
    assert 'uint8' in str(ex.value)


@pytest.mark.parametrize('build_ovw', [True, False])
def test_process_overview(build_ovw, rgb_pinhole_utm34n_ortho: Ortho, tmp_path: Path):
    """Test overview(s) are created according to the ``build_ovw`` value."""
    ortho_file = tmp_path.joinpath('test_ortho.tif')
    rgb_pinhole_utm34n_ortho.process(ortho_file, (0.25, 0.25), build_ovw=build_ovw)
    assert ortho_file.exists()

    with rio.open(ortho_file, 'r') as ortho_im:
        assert min(ortho_im.shape) >= 512
        if build_ovw:
            assert len(ortho_im.overviews(1)) > 0
        else:
            assert len(ortho_im.overviews(1)) == 0


@pytest.mark.parametrize(
    'camera',
    ['pinhole_camera', 'brown_camera', 'opencv_camera', 'fisheye_camera', 'rpc_camera_proj'],
)
def test_process_camera(
    rgb_byte_src_file: Path,
    float_utm34n_dem_file: Path,
    camera: str,
    utm34n_crs,
    tmp_path: Path,
    request: pytest.FixtureRequest,
):
    """Test ortho image format and content for different cameras."""
    # create Ortho object and process, using the planar DEM band
    camera: Camera = request.getfixturevalue(camera)
    ortho = Ortho(rgb_byte_src_file, float_utm34n_dem_file, camera, utm34n_crs, dem_band=2)
    resolution = (5, 5)
    ortho_file = tmp_path.joinpath('test_ortho.tif')
    ortho.process(ortho_file, resolution, compress=Compress.deflate, interp=Interp.nearest)
    dem_bounds = np.array(array_bounds(*ortho._dem_array.shape, ortho._dem_transform))
    assert ortho_file.exists()

    with rio.open(rgb_byte_src_file, 'r') as src_im, rio.open(ortho_file, 'r') as ortho_im:
        # test ortho bounds and content
        ortho_bounds = np.array(ortho_im.bounds)
        assert np.all(ortho_bounds[:2] >= dem_bounds[:2]) and np.all(
            ortho_bounds[-2:] <= dem_bounds[-2:]
        )

        src_array = src_im.read()
        ortho_array = ortho_im.read()
        ortho_mask = ortho_im.dataset_mask().astype('bool')
        assert np.all(np.unique(src_array) == np.unique(ortho_array[:, ortho_mask]))
        assert src_array.mean() == pytest.approx(ortho_array[:, ortho_mask].mean(), abs=15)
        assert src_array.std() == pytest.approx(ortho_array[:, ortho_mask].std(), abs=15)


def test_process_progress(
    rgb_pinhole_utm34n_ortho: Ortho, tmp_path: Path, capsys: pytest.CaptureFixture
):
    """Test ortho progress bar display."""
    # default bar
    ortho_file = tmp_path.joinpath('test_ortho.tif')
    rgb_pinhole_utm34n_ortho.process(ortho_file, _dem_resolution, progress=True)
    cap = capsys.readouterr()
    assert 'blocks' in cap.err and '100%' in cap.err

    # no bar
    rgb_pinhole_utm34n_ortho.process(ortho_file, _dem_resolution, overwrite=True, progress=False)
    cap = capsys.readouterr()
    assert 'blocks' not in cap.err and '100%' not in cap.err

    # custom bar
    desc = 'custom'
    rgb_pinhole_utm34n_ortho.process(
        ortho_file, _dem_resolution, overwrite=True, progress=dict(desc=desc)
    )
    cap = capsys.readouterr()
    assert desc in cap.err


def test_process_ngi(
    ngi_image_files: tuple[Path, ...],
    ngi_dem_file: Path,
    ngi_legacy_config_file: Path,
    ngi_legacy_csv_file: Path,
    ngi_crs: str,
    tmp_path: Path,
):
    """Test integration and ortho overlap using NGI aerial images."""
    int_param_dict = param_io.read_oty_int_param(ngi_legacy_config_file)
    ext_param_dict = param_io.CsvReader(ngi_legacy_csv_file, crs=ngi_crs).read_ext_param()
    camera = create_camera(**next(iter(int_param_dict.values())))

    ortho_files = []
    for src_file in ngi_image_files:
        ortho_file = tmp_path.joinpath(src_file.stem + '_ORTHO.tif')
        ext_param = ext_param_dict[src_file.stem]
        camera.update(xyz=ext_param['xyz'], opk=ext_param['opk'])
        ortho = Ortho(src_file, ngi_dem_file, camera, crs=ngi_crs)
        ortho.process(ortho_file, resolution=(5, 5))

        assert ortho_file.exists()
        ortho_files.append(ortho_file)

    _validate_ortho_files(tuple(ortho_files))


def test_process_odm(
    odm_image_files: tuple[Path, ...],
    odm_dem_file: Path,
    odm_reconstruction_file: Path,
    odm_crs: str,
    tmp_path: Path,
):
    """Test integration and ortho overlap using ODM drone images."""
    reader = param_io.OsfmReader(odm_reconstruction_file, crs=odm_crs)
    int_param_dict = reader.read_int_param()
    ext_param_dict = reader.read_ext_param()
    camera = create_camera(**next(iter(int_param_dict.values())))

    ortho_files = []
    for src_file in odm_image_files:
        ortho_file = tmp_path.joinpath(src_file.stem + '_ORTHO.tif')
        ext_param = ext_param_dict[src_file.stem]
        camera.update(xyz=ext_param['xyz'], opk=ext_param['opk'])
        ortho = Ortho(src_file, odm_dem_file, camera, crs=odm_crs)
        ortho.process(ortho_file, resolution=(0.25, 0.25))

        assert ortho_file.exists()
        ortho_files.append(ortho_file)

    _validate_ortho_files(tuple(ortho_files), num_ovl_thresh=5)



# Copyright The Orthority Contributors.
#
# This file is part of Orthority.
#
# Orthority is free software: you can redistribute it and/or modify it under the terms of the GNU
# Affero General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# Orthority is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Orthority.
# If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from contextlib import contextmanager
from io import TextIOWrapper
from pathlib import Path

import fsspec
import numpy as np
import pytest
import rasterio as rio

from orthority import utils


@pytest.mark.parametrize('file', ['odm_image_file', 'odm_image_url'])
def test_get_filename(file: str, request: pytest.FixtureRequest):
    """Test get_filename() returns with different ``file`` objects."""
    file = request.getfixturevalue(file)
    exp_val = Path(file).name
    assert utils.get_filename(file) == exp_val
    ofile = fsspec.open(file, 'rb')
    assert utils.get_filename(ofile) == exp_val
    with utils.Open(file, 'rb') as f:
        assert utils.get_filename(f) == exp_val
    with utils.OpenRaster(file, 'r') as f:
        assert utils.get_filename(f) == exp_val


@pytest.mark.parametrize('file', ['odm_image_file', 'odm_image_url'])
def test_join_ofile(file: str, request: pytest.FixtureRequest):
    """Test join_ofile() returns valid OpenFile instances for existing paths / URIs,
    with different base path / URI types, and with and without trailing slashes on the base path
    / URI.
    """
    file = request.getfixturevalue(file)
    ofile = fsspec.open(file)
    parts = str(file).replace('\\', '/').split('/')

    path_types = [str, fsspec.open, Path] if isinstance(file, Path) else [str, fsspec.open]
    for path_type in path_types:
        for pidx in [1, 2]:
            _base_path = '/'.join(parts[:-pidx])
            rel_path = '/'.join(parts[-pidx:])

            for base_path in map(path_type, [_base_path, _base_path + '/']):
                join_ofile = utils.join_ofile(base_path, rel_path)
                assert isinstance(join_ofile, fsspec.core.OpenFile)
                assert join_ofile.path == ofile.path
                assert join_ofile.fs.exists(join_ofile.path)


@pytest.mark.parametrize('raster_file', ['odm_image_file', 'odm_image_url'])
def test_open_raster_read(raster_file: str, request: pytest.FixtureRequest):
    """Test OpenRaster in ``'r'`` mode with path / URI, fsspec OpenFile and dataset objects."""
    raster_file = request.getfixturevalue(raster_file)

    def _test_open_raster_read(_raster_file, test_closed: bool = True):
        """Test OpenRaster for the given ``_raster_file`` object."""
        with utils.OpenRaster(_raster_file, 'r') as _ds:
            assert isinstance(_ds, rio.DatasetReader)
            assert not _ds.closed
            assert _ds.filename == Path(raster_file).name
            _ = _ds.read(1, window=_ds.block_window(1, 0, 0))
        if test_closed:
            assert _ds.closed

    _test_open_raster_read(raster_file)
    _test_open_raster_read(fsspec.open(raster_file, 'rb'))
    with utils.OpenRaster(raster_file, 'r') as ds:
        _test_open_raster_read(ds, test_closed=False)
    assert ds.closed


def test_open_raster_write(tmp_path: Path):
    """Test OpenRaster in ``'w'`` mode with path / URI, fsspec OpenFile and dataset objects."""
    raster_file = tmp_path.joinpath('test_open_raster_write.tif')
    profile = rio.default_gtiff_profile
    array = np.ones((256, 256), dtype=profile['dtype'])
    profile.update(width=array.shape[1], height=array.shape[0], count=1)

    @contextmanager
    def _test_temp_file(filename: Path):
        """Test ``filename`` exists on exit, then delete."""
        try:
            yield filename
        finally:
            assert filename.exists()
            filename.unlink()

    def _test_open_raster_write(_raster_file, test_closed: bool = True):
        """Test OpenRaster for the given ``_raster_file`` object."""
        with utils.OpenRaster(_raster_file, 'w', **profile) as _ds:
            assert isinstance(_ds, rio.io.DatasetWriter)
            assert not _ds.closed
            assert _ds.filename == Path(raster_file).name
            _ds.write(array, indexes=1)
        if test_closed:
            assert _ds.closed

    with _test_temp_file(raster_file):
        _test_open_raster_write(raster_file)
    with _test_temp_file(raster_file):
        _test_open_raster_write(fsspec.open(str(raster_file), 'wb'))
    with _test_temp_file(raster_file):
        with utils.OpenRaster(raster_file, 'w', **profile) as ds:
            _test_open_raster_write(ds, test_closed=False)
        assert ds.closed


def test_open_raster_overwrite(tmp_path: Path):
    """Test the OpenRaster ``overwrite`` argument."""
    raster_file = tmp_path.joinpath('test_open_raster_overwrite.tif')
    profile = rio.default_gtiff_profile
    array = np.ones((256, 256), dtype=profile['dtype'])
    profile.update(width=array.shape[1], height=array.shape[0], count=1)
    raster_file.touch()

    def _test_open_raster_overwrite(_raster_file, overwrite: bool):
        with utils.OpenRaster(_raster_file, 'w', overwrite=overwrite, **profile) as _ds:
            _ds.write(array, indexes=1)

    # test overwriting an existing file with overwrite=True
    _test_open_raster_overwrite(raster_file, overwrite=True)
    _test_open_raster_overwrite(fsspec.open(str(raster_file), 'wb'), overwrite=True)

    # test writing to an existing file with overwrite=False raises FileExistsError
    with pytest.raises(FileExistsError) as ex:
        _test_open_raster_overwrite(raster_file, overwrite=False)
    assert raster_file.name in str(ex.value)
    with pytest.raises(FileExistsError) as ex:
        _test_open_raster_overwrite(fsspec.open(str(raster_file), 'wb'), overwrite=False)
    assert raster_file.name in str(ex.value)


@pytest.mark.parametrize(
    'raster_file',
    ['unknown/unknown.tif', 'https://un.known/unknown.tif', 'https://github.com/unknown.tif'],
)
def test_open_raster_not_found_error(raster_file: str):
    """Test OpenRaster raises a FileNotFoundError error with non-existing file path / URIs."""
    with pytest.raises(FileNotFoundError):
        with utils.OpenRaster(raster_file, 'r'):
            pass
    ofile = fsspec.open(raster_file, 'rb')
    with pytest.raises(FileNotFoundError):
        with utils.OpenRaster(ofile, 'r'):
            pass


@pytest.mark.parametrize(
    'file, kwargs',
    [
        ('ngi_oty_int_param_file', dict(newline=None, encoding='utf8')),
        ('ngi_oty_int_param_url', dict()),
    ],
)
def test_open_read(file: str, kwargs: dict, request: pytest.FixtureRequest):
    """Test Open in ``'rt'`` mode with path / URI, fsspec OpenFile and file objects."""
    file = request.getfixturevalue(file)

    def _test_open_read(_file, test_closed: bool = True, **kwargs):
        """Test Open for the given ``_file`` object."""
        with utils.Open(_file, 'rt', **kwargs) as _f:
            assert isinstance(_f, TextIOWrapper)
            assert not _f.closed
            assert _f.filename == Path(file).name
            _ = _f.read()
        if test_closed:
            assert _f.closed

    _test_open_read(file)
    _test_open_read(fsspec.open(file, 'rt'))
    with utils.Open(file, 'rt', **kwargs) as f:
        _test_open_read(f, test_closed=False, **kwargs)
    assert f.closed


def test_open_write(tmp_path: Path):
    """Test Open in ``'w'`` mode with path / URI, fsspec OpenFile and file objects."""
    file = tmp_path.joinpath('test_open_write.txt')

    @contextmanager
    def _test_temp_file(filename: Path):
        """Test ``filename`` exists on exit, then delete."""
        try:
            yield filename
        finally:
            assert filename.exists()
            assert filename.read_text()
            filename.unlink()

    def _test_open_write(_file, test_closed: bool = True, **kwargs):
        """Test Open for the given ``_file`` object."""
        with utils.Open(_file, 'wt', **kwargs) as _f:
            assert isinstance(_f, TextIOWrapper)
            assert not _f.closed
            assert _f.filename == Path(file).name
            _f.write('test')
        if test_closed:
            assert _f.closed

    kwargs = dict(newline=None, encoding='utf8')
    with _test_temp_file(file):
        _test_open_write(file, **kwargs)
    with _test_temp_file(file):
        _test_open_write(fsspec.open(str(file), 'wt'), **kwargs)
    with _test_temp_file(file):
        with utils.Open(file, 'wt', **kwargs) as f:
            _test_open_write(f, test_closed=False, **kwargs)
        assert f.closed


def test_open_overwrite(tmp_path: Path):
    """Test the Open ``overwrite`` argument."""
    file = tmp_path.joinpath('test_open_overwrite.txt')
    file.touch()

    def _test_open_overwrite(_file, overwrite: bool):
        with utils.Open(_file, 'wt', overwrite=overwrite) as _f:
            _f.write('test')

    # test overwriting an existing file with overwrite=True
    _test_open_overwrite(file, overwrite=True)
    _test_open_overwrite(fsspec.open(str(file), 'wt'), overwrite=True)

    # test writing to an existing file with overwrite=False raises FileExistsError
    with pytest.raises(FileExistsError) as ex:
        _test_open_overwrite(file, overwrite=False)
    assert file.name in str(ex.value)
    with pytest.raises(FileExistsError) as ex:
        _test_open_overwrite(fsspec.open(str(file), 'wt'), overwrite=False)
    assert file.name in str(ex.value)


@pytest.mark.parametrize(
    'file',
    ['unknown/unknown.txt', 'https://un.known/unknown.txt', 'https://github.com/unknown.txt'],
)
def test_open_not_found_error(file: str):
    """Test Open raises a FileNotFoundError error with non-existing file path / URIs."""
    with pytest.raises(FileNotFoundError):
        with utils.Open(file, 'rt'):
            pass
    ofile = fsspec.open(file, 'rt')
    with pytest.raises(FileNotFoundError):
        with utils.Open(ofile, 'rt'):
            pass



# Copyright The Orthority Contributors.
#
# This file is part of Orthority.
#
# Orthority is free software: you can redistribute it and/or modify it under the terms of the GNU
# Affero General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# Orthority is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Orthority.
# If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from enum import Enum

import cv2
from rasterio.enums import Resampling


class CameraType(str, Enum):
    """Camera model types."""

    pinhole = 'pinhole'
    """Pinhole frame camera model."""

    brown = 'brown'
    """
    Brown-Conrady frame camera model.

    Compatible with `OpenDroneMap / OpenSfM 
    <https://opensfm.org/docs/geometry.html#camera-models>`__ ``perspective``, ``simple_radial``, 
    ``radial`` and ``brown`` model parameters, and the 4- and 5-coefficient versions of the 
    `OpenCV general model <https://docs.opencv.org/4.x/d9/d0c/group__calib3d.html>`__."""

    fisheye = 'fisheye'
    """
    Fisheye frame camera model.

    Compatible with `OpenDroneMap / OpenSfM 
    <https://opensfm.org/docs/geometry.html#fisheye-camera>`__ ``fisheye``, and `OpenCV 
    <https://docs.opencv.org/4.x/db/d58/group__calib3d__fisheye.html>`__  fisheye model 
    parameters."""

    opencv = 'opencv'
    """
    `OpenCV general frame camera model <https://docs.opencv.org/4.x/d9/d0c/group__calib3d.html>`__.
    """

    rpc = 'rpc'
    """RPC camera model."""

    def __repr__(self):
        return self._name_

    def __str__(self):
        return self._name_

    @classmethod
    def from_odm(cls, cam_type: str):
        """Convert from OpenDroneMap / OpenSfM projection type."""
        cam_type = cam_type.lower().strip()
        if cam_type in ['perspective', 'simple_radial', 'radial']:
            cam_type = 'brown'
        if cam_type not in cls.__members__ or cam_type == 'rpc':
            raise ValueError(f"Unsupported OpenDroneMap / OpenSfM camera type: '{cam_type}'")
        return cls(cam_type)


class Interp(str, Enum):
    """Interpolation types."""

    nearest = 'nearest'
    """Nearest neighbor interpolation."""
    average = 'average'
    """Average input pixels over the corresponding output pixel area (suited to downsampling)."""
    bilinear = 'bilinear'
    """Bilinear interpolation."""
    cubic = 'cubic'
    """Bicubic interpolation."""
    lanczos = 'lanczos'
    """Lanczos windowed sinc interpolation."""

    def __repr__(self):
        return self._name_

    def __str__(self):
        return self._name_

    def to_cv(self) -> int:
        """Convert to OpenCV interpolation type."""
        name_to_cv = dict(
            average=cv2.INTER_AREA,
            bilinear=cv2.INTER_LINEAR,
            cubic=cv2.INTER_CUBIC,
            lanczos=cv2.INTER_LANCZOS4,
            nearest=cv2.INTER_NEAREST,
        )
        return name_to_cv[self._name_]

    def to_rio(self) -> Resampling:
        """Convert to rasterio resampling type."""
        return Resampling[self._name_]


class Compress(str, Enum):
    """Compression types."""

    jpeg = 'jpeg'
    """JPEG compression."""
    deflate = 'deflate'
    """Deflate compression."""

    def __repr__(self):
        return self._name_

    def __str__(self):
        return self._name_


class CsvFormat(Enum):
    """Type of the position and orientation values in a CSV exterior parameter file."""

    xyz_opk = 1
    """Projected (x, y, z) position and (omega, phi, kappa) orientation."""
    xyz_rpy = 2
    """Projected (x, y, z) position and (roll, pitch, yaw) orientation."""
    lla_opk = 3
    """Geographic (latitude, longitude, altitude) position and (omega, phi, kappa) orientation."""
    lla_rpy = 4
    """Geographic (latitude, longitude, altitude) position and (roll, pitch, yaw) orientation."""

    @property
    def is_xyz(self) -> bool:
        """Whether the format has (x, y, z) position."""
        return self is CsvFormat.xyz_opk or self is CsvFormat.xyz_rpy

    @property
    def is_opk(self) -> bool:
        """Whether the format has (omega, phi, kappa) orientation."""
        return self is CsvFormat.xyz_opk or self is CsvFormat.lla_opk
